# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_errors import validate_transaction_errors

from newrelic.packages import six
from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics


def target_application(with_tweens=False, tweens_explicit=False):
    # We need to delay Pyramid application creation because of ordering
    # issues whereby the agent needs to be initialised before Pyramid is
    # imported and the routes configured. Normally pytest only runs the
    # global fixture which will initialise the agent after each test
    # file is imported, which is too late. We also can't do application
    # creation within a function as Pyramid relies on view handlers being
    # at global scope, so import it from a separate module.

    from _test_application import target_application as _app
    return _app(with_tweens, tweens_explicit)


if six.PY3:
    tween_name = ('Function/_test_application:'
                  'simple_tween_factory.<locals>.simple_tween')
else:
    tween_name = 'Function/_test_application:simple_tween'

_test_application_index_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/pyramid.router:Router.__call__', 1),
        ('Function/_test_application:home_view', 1)]


@pytest.mark.parametrize('with_tweens,tweens_explicit', (
    (False, False),
    (True, False),
    (True, True),
))
def test_application_index(with_tweens, tweens_explicit):
    application = target_application(with_tweens, tweens_explicit)

    metrics = list(_test_application_index_scoped_metrics)
    if with_tweens:
        metrics.append((tween_name, 1))

    @validate_code_level_metrics("_test_application", "home_view")
    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        '_test_application:home_view', scoped_metrics=metrics)
    def _test():
        response = application.get('')
        response.mustcontain('INDEX RESPONSE')

    _test()


_test_not_found_as_exception_response_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/pyramid.router:Router.__call__', 1),
        ('Function/pyramid.httpexceptions:default_exceptionresponse_view', 1),
        ('Function/_test_application:not_found_exception_response', 1)]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    '_test_application:not_found_exception_response',
    scoped_metrics=_test_not_found_as_exception_response_scoped_metrics)
def test_not_found_as_exception_response():
    application = target_application()
    application.get('/nf1', status=404)


_test_not_found_raises_NotFound_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/pyramid.router:Router.__call__', 1),
        ('Function/pyramid.httpexceptions:default_exceptionresponse_view', 1),
        ('Function/_test_application:raise_not_found', 1)]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    '_test_application:raise_not_found',
    scoped_metrics=_test_not_found_raises_NotFound_scoped_metrics)
def test_application_not_found_raises_NotFound():
    application = target_application()
    application.get('/nf2', status=404)


_test_not_found_returns_NotFound_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/pyramid.router:Router.__call__', 1),
        ('Function/_test_application:return_not_found', 1)]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    '_test_application:return_not_found',
    scoped_metrics=_test_not_found_returns_NotFound_scoped_metrics)
def test_application_not_found_returns_NotFound():
    application = target_application()
    application.get('/nf3', status=404)


_test_unexpected_exception_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Function/pyramid.router:Router.__call__', 1),
        ('Function/_test_application:error', 1)]

if six.PY3:
    _test_unexpected_exception_errors = ['builtins:RuntimeError']
else:
    _test_unexpected_exception_errors = ['exceptions:RuntimeError']


@validate_transaction_errors(errors=_test_unexpected_exception_errors)
@validate_transaction_metrics(
    '_test_application:error',
    scoped_metrics=_test_unexpected_exception_scoped_metrics)
def test_application_unexpected_exception():
    application = target_application()
    with pytest.raises(RuntimeError):
        application.get('/error', status=500)


_test_redirect_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/pyramid.router:Router.__call__', 1),
        ('Function/pyramid.httpexceptions:default_exceptionresponse_view', 1),
        ('Function/_test_application:redirect', 1)]


@validate_code_level_metrics("_test_application", "redirect")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    '_test_application:redirect',
    scoped_metrics=_test_redirect_scoped_metrics)
def test_application_redirect():
    application = target_application()
    application.get('/redirect', status=302)


_test_resource_get_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/pyramid.router:Router.__call__', 1),
        ('Function/_test_application:RestView', 1),
        ('Function/_test_application:RestView.get', 1)]


@validate_code_level_metrics("_test_application.RestView", "get")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    '_test_application:RestView.get',
    scoped_metrics=_test_resource_get_scoped_metrics)
def test_application_rest_get():
    application = target_application()
    response = application.get('/rest')
    response.mustcontain('Called GET')


_test_resource_post_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/pyramid.router:Router.__call__', 1),
        ('Function/_test_application:RestView', 2),
        ('Function/_test_application:RestView.post', 1)]


@validate_code_level_metrics("_test_application.RestView", "post")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    '_test_application:RestView.post',
    scoped_metrics=_test_resource_post_scoped_metrics)
def test_application_rest_post():
    application = target_application()
    response = application.post('/rest')  # Raises PredicateMismatch
    response.mustcontain('Called POST')


_test_html_insertion_settings = {
    'browser_monitoring.enabled': True,
    'browser_monitoring.auto_instrument': True,
    'js_agent_loader': u'<!-- NREUM HEADER -->',
}


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion():
    application = target_application()
    response = application.get('/html_insertion', status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    response.mustcontain('NREUM HEADER', 'NREUM.info')



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# import aioredis
# import pytest
# from conftest import AIOREDIS_VERSION, event_loop
from testing_support.db_settings import redis_settings
from testing_support.fixtures import override_application_settings
from testing_support.util import instance_hostname
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

from newrelic.api.background_task import background_task

DB_SETTINGS = redis_settings()[0]

_enable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": True,
    "datastore_tracer.database_name_reporting.enabled": True,
}
_disable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": False,
    "datastore_tracer.database_name_reporting.enabled": False,
}
_instance_only_settings = {
    "datastore_tracer.instance_reporting.enabled": True,
    "datastore_tracer.database_name_reporting.enabled": False,
}
_database_only_settings = {
    "datastore_tracer.instance_reporting.enabled": False,
    "datastore_tracer.database_name_reporting.enabled": True,
}


_enabled_required = {
    "host": instance_hostname(DB_SETTINGS["host"]),
    "port_path_or_id": str(DB_SETTINGS["port"]),
    "db.instance": "0",
}
_enabled_forgone = {}

_disabled_required = {}
_disabled_forgone = {
    "host": "VALUE NOT USED",
    "port_path_or_id": "VALUE NOT USED",
    "db.instance": "VALUE NOT USED",
}


_instance_only_required = {
    "host": instance_hostname(DB_SETTINGS["host"]),
    "port_path_or_id": str(DB_SETTINGS["port"]),
}
_instance_only_forgone = {
    "db.instance": "0",
}

_database_only_required = {
    "db.instance": "0",
}
_database_only_forgone = {
    "host": "VALUE NOT USED",
    "port_path_or_id": "VALUE NOT USED",
}


async def exercise_redis(client):
    await client.set("key", "value")
    await client.get("key")

    if hasattr(client, "execute_command"):
        await client.execute_command("CLIENT", "LIST", parse="LIST")
    else:
        await client.execute("CLIENT", "LIST")


@override_application_settings(_enable_instance_settings)
@validate_tt_collector_json(datastore_params=_enabled_required, datastore_forgone_params=_enabled_forgone)
@background_task()
def test_trace_node_datastore_params_enable_instance(client, loop):
    loop.run_until_complete(exercise_redis(client))


@override_application_settings(_disable_instance_settings)
@validate_tt_collector_json(datastore_params=_disabled_required, datastore_forgone_params=_disabled_forgone)
@background_task()
def test_trace_node_datastore_params_disable_instance(client, loop):
    loop.run_until_complete(exercise_redis(client))


@override_application_settings(_instance_only_settings)
@validate_tt_collector_json(datastore_params=_instance_only_required, datastore_forgone_params=_instance_only_forgone)
@background_task()
def test_trace_node_datastore_params_instance_only(client, loop):
    loop.run_until_complete(exercise_redis(client))


@override_application_settings(_database_only_settings)
@validate_tt_collector_json(datastore_params=_database_only_required, datastore_forgone_params=_database_only_forgone)
@background_task()
def test_trace_node_datastore_params_database_only(client, loop):
    loop.run_until_complete(exercise_redis(client))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc
import os
import platform

import pytest
from testing_support.fixtures import override_generic_settings

from newrelic.core.config import global_settings
from newrelic.packages import six
from newrelic.samplers.cpu_usage import cpu_usage_data_source
from newrelic.samplers.gc_data import garbage_collector_data_source
from newrelic.samplers.memory_usage import memory_usage_data_source

settings = global_settings()


@pytest.fixture
def gc_data_source():
    sampler = garbage_collector_data_source(settings=())["factory"](environ=())
    sampler.start()
    yield sampler
    sampler.stop()


@pytest.fixture
def cpu_data_source():
    sampler = cpu_usage_data_source(settings=())["factory"](environ=())
    sampler.start()
    yield sampler
    sampler.stop()


@pytest.fixture
def memory_data_source():
    sampler = memory_usage_data_source(settings=())["factory"](environ=())
    yield sampler


PID = os.getpid()

if six.PY2:
    EXPECTED_GC_METRICS = (
        "GC/objects/%d/all" % PID,
        "GC/objects/%d/generation/0" % PID,
        "GC/objects/%d/generation/1" % PID,
        "GC/objects/%d/generation/2" % PID,
    )
else:
    EXPECTED_GC_METRICS = (
        "GC/objects/%d/all" % PID,
        "GC/objects/%d/generation/0" % PID,
        "GC/objects/%d/generation/1" % PID,
        "GC/objects/%d/generation/2" % PID,
        "GC/collections/%d/all" % PID,
        "GC/collections/%d/0" % PID,
        "GC/collections/%d/1" % PID,
        "GC/collections/%d/2" % PID,
        "GC/collected/%d/all" % PID,
        "GC/collected/%d/0" % PID,
        "GC/collected/%d/1" % PID,
        "GC/collected/%d/2" % PID,
        "GC/uncollectable/%d/all" % PID,
        "GC/uncollectable/%d/0" % PID,
        "GC/uncollectable/%d/1" % PID,
        "GC/uncollectable/%d/2" % PID,
        "GC/time/%d/all" % PID,
        "GC/time/%d/0" % PID,
        "GC/time/%d/1" % PID,
        "GC/time/%d/2" % PID,
    )


@pytest.mark.xfail(
    platform.python_implementation() == "PyPy",
    reason="Not implemented on PyPy yet",
    strict=True,
    raises=AssertionError,
)
@pytest.mark.parametrize("top_object_count_limit", (1, 0))
def test_gc_metrics_collection(gc_data_source, top_object_count_limit):
    @override_generic_settings(
        settings,
        {
            "gc_runtime_metrics.enabled": True,
            "gc_runtime_metrics.top_object_count_limit": top_object_count_limit,
        },
    )
    def _test():
        gc.collect()
        metrics_table = set(m[0] for m in (gc_data_source() or ()))

        for metric in EXPECTED_GC_METRICS:
            assert metric in metrics_table

        # Verify object count by type metrics are recorded
        obj_metric_count = 0
        for metric in metrics_table:
            if metric.startswith("GC/objects/"):
                obj_metric_count += 1

        if top_object_count_limit > 0:
            assert obj_metric_count == 5, metrics_table
        else:
            assert obj_metric_count == 4, metrics_table

    _test()


@pytest.mark.skipif(
    platform.python_implementation() == "PyPy",
    reason="GC Metrics are always disabled on PyPy",
)
@pytest.mark.parametrize("enabled", (True, False))
def test_gc_metrics_config(gc_data_source, enabled):
    @override_generic_settings(settings, {"gc_runtime_metrics.enabled": enabled})
    def _test():
        assert gc_data_source.enabled == enabled

    _test()


EXPECTED_CPU_METRICS = (
    "CPU/User Time",
    "CPU/User/Utilization",
    "CPU/System Time",
    "CPU/System/Utilization",
    "CPU/Total Time",
    "CPU/Total/Utilization",
)


def test_cpu_metrics_collection(cpu_data_source):
    metrics_table = set(m[0] for m in (cpu_data_source() or ()))

    for metric in EXPECTED_CPU_METRICS:
        assert metric in metrics_table


EXPECTED_MEMORY_METRICS = (
    "Memory/Physical",
    "Memory/Physical/%d" % PID,
    "Memory/Physical/Utilization",
    "Memory/Physical/Utilization/%d" % PID,
)


def test_memory_metrics_collection(memory_data_source):
    metrics_table = set(m[0] for m in (memory_data_source() or ()))

    for metric in EXPECTED_MEMORY_METRICS:
        assert metric in metrics_table



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from conftest import CONDITION_CLS

from newrelic.common.streaming_utils import StreamBuffer, StreamBufferIterator
from newrelic.core.infinite_tracing_pb2 import Span, SpanBatch


class StopIterationOnWait(CONDITION_CLS):
    def wait(self, *args, **kwargs):
        raise StopIteration()


@staticmethod
def stop_iteration_condition(*args, **kwargs):
    return StopIterationOnWait(*args, **kwargs)


@pytest.fixture(scope="function")
def stop_iteration_on_wait(monkeypatch):
    monkeypatch.setattr(StreamBuffer, "condition", stop_iteration_condition)


def test_stream_buffer_iterator_batching(stop_iteration_on_wait, batching):
    stream_buffer = StreamBuffer(5, batching=batching)

    for _ in range(5):
        span = Span(intrinsics={}, agent_attributes={}, user_attributes={})
        stream_buffer.put(span)

    buffer_contents = list(stream_buffer)
    if batching:
        assert len(buffer_contents) == 1
        assert isinstance(buffer_contents.pop(), SpanBatch)
    else:
        assert len(buffer_contents) == 5
        assert all(isinstance(element, Span) for element in stream_buffer)


def test_stream_buffer_iterator_max_batch_size(stop_iteration_on_wait):
    stream_buffer = StreamBuffer(StreamBufferIterator.MAX_BATCH_SIZE + 1, batching=True)

    # Add 1 more span than the maximum batch size
    for _ in range(StreamBufferIterator.MAX_BATCH_SIZE + 1):
        span = Span(intrinsics={}, agent_attributes={}, user_attributes={})
        stream_buffer.put(span)

    # Pull all span batches out of buffer
    buffer_contents = list(stream_buffer)
    assert len(buffer_contents) == 2

    # Large batch
    batch = buffer_contents.pop(0)
    assert isinstance(batch, SpanBatch)
    assert len(batch.spans) == StreamBufferIterator.MAX_BATCH_SIZE

    # Single span batch
    batch = buffer_contents.pop(0)
    assert isinstance(batch, SpanBatch)
    assert len(batch.spans) == 1


def test_stream_buffer_queue_size():
    stream_buffer = StreamBuffer(1)

    # Add more spans than queue can hold
    for _ in range(2):
        span = Span(intrinsics={}, agent_attributes={}, user_attributes={})
        stream_buffer.put(span)

    # Ensure spans are dropped and not stored
    assert len(stream_buffer) == 1
    assert stream_buffer._dropped == 1
    assert stream_buffer._seen == 2



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import pytest
from testing_support.asgi_testing import AsgiTest
from testing_support.fixtures import override_application_settings
from testing_support.sample_asgi_applications import (
    AppWithDescriptor,
    simple_app_v2,
    simple_app_v2_init_exc,
    simple_app_v2_raw,
    simple_app_v3,
    simple_app_v3_raw,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.asgi_application import ASGIApplicationWrapper, asgi_application

# Setup test apps from sample_asgi_applications.py
simple_app_v2_original = AsgiTest(simple_app_v2_raw)
simple_app_v3_original = AsgiTest(simple_app_v3_raw)

simple_app_v3_wrapped = AsgiTest(simple_app_v3)
simple_app_v2_wrapped = AsgiTest(simple_app_v2)
simple_app_v2_init_exc = AsgiTest(simple_app_v2_init_exc)


# Test naming scheme logic and ASGIApplicationWrapper for a single callable
@pytest.mark.parametrize("naming_scheme", (None, "component", "framework"))
def test_single_callable_naming_scheme(naming_scheme):

    if naming_scheme in ("component", "framework"):
        expected_name = "testing_support.sample_asgi_applications:simple_app_v3_raw"
        expected_group = "Function"
    else:
        expected_name = ""
        expected_group = "Uri"

    @validate_transaction_metrics(name=expected_name, group=expected_group)
    @override_application_settings({"transaction_name.naming_scheme": naming_scheme})
    def _test():
        response = simple_app_v3_wrapped.make_request("GET", "/")
        assert response.status == 200
        assert response.headers == {}
        assert response.body == b""

    _test()


# Test the default naming scheme logic and ASGIApplicationWrapper for a double callable
@validate_transaction_metrics(name="", group="Uri")
def test_double_callable_default_naming_scheme():
    response = simple_app_v2_wrapped.make_request("GET", "/")
    assert response.status == 200
    assert response.headers == {}
    assert response.body == b""


# No harm test on single callable asgi app with agent disabled to ensure proper response
def test_single_callable_raw():
    response = simple_app_v3_original.make_request("GET", "/")
    assert response.status == 200
    assert response.headers == {}
    assert response.body == b""


# No harm test on double callable asgi app with agent disabled to ensure proper response
def test_double_callable_raw():
    response = simple_app_v2_original.make_request("GET", "/")
    assert response.status == 200
    assert response.headers == {}
    assert response.body == b""


# Test asgi_application decorator with parameters passed in on a single callable
@pytest.mark.parametrize("name, group", ((None, "group"), ("name", "group"), ("", "group")))
def test_asgi_application_decorator_single_callable(name, group):
    if name:
        expected_name = name
        expected_group = group
    else:
        expected_name = ""
        expected_group = "Uri"

    @validate_transaction_metrics(name=expected_name, group=expected_group)
    def _test():
        asgi_decorator = asgi_application(name=name, group=group)
        decorated_application = asgi_decorator(simple_app_v3_raw)
        application = AsgiTest(decorated_application)
        response = application.make_request("GET", "/")
        assert response.status == 200
        assert response.headers == {}
        assert response.body == b""

    _test()


# Test asgi_application decorator using default values on a double callable
@validate_transaction_metrics(name="", group="Uri")
def test_asgi_application_decorator_no_params_double_callable():
    asgi_decorator = asgi_application()
    decorated_application = asgi_decorator(simple_app_v2_raw)
    application = AsgiTest(decorated_application)
    response = application.make_request("GET", "/")
    assert response.status == 200
    assert response.headers == {}
    assert response.body == b""


# Test for presence of framework and dispatcher info based on whether framework is specified
@validate_transaction_metrics(
    name="test", custom_metrics=[("Python/Framework/framework/v1", 1), ("Python/Dispatcher/dispatcher/v1.0.0", 1)]
)
def test_dispatcher_and_framework_metrics():
    asgi_decorator = asgi_application(name="test", framework=("framework", "v1"), dispatcher=("dispatcher", "v1.0.0"))
    decorated_application = asgi_decorator(simple_app_v2_raw)
    application = AsgiTest(decorated_application)
    application.make_request("GET", "/")


# Test for presence of framework and dispatcher info under existing transaction
@validate_transaction_metrics(
    name="test", custom_metrics=[("Python/Framework/framework/v1", 1), ("Python/Dispatcher/dispatcher/v1.0.0", 1)]
)
def test_double_wrapped_dispatcher_and_framework_metrics():
    inner_asgi_decorator = asgi_application(
        name="test", framework=("framework", "v1"), dispatcher=("dispatcher", "v1.0.0")
    )
    decorated_application = inner_asgi_decorator(simple_app_v2_raw)

    outer_asgi_decorator = asgi_application(name="double_wrapped")
    double_decorated_application = outer_asgi_decorator(decorated_application)

    application = AsgiTest(double_decorated_application)
    application.make_request("GET", "/")


@pytest.mark.parametrize("method", ("method", "cls", "static"))
@validate_transaction_metrics(name="", group="Uri")
def test_app_with_descriptor(method):
    application = AsgiTest(getattr(AppWithDescriptor(), method))
    response = application.make_request("GET", "/")
    assert response.status == 200
    assert response.headers == {}
    assert response.body == b""


# Verify that errors are not generated when using multiple wrappers
def test_multiple_calls_to_asgi_wrapper(caplog):
    app = ASGIApplicationWrapper(simple_app_v3)
    app = AsgiTest(app)

    with caplog.at_level(logging.ERROR, logger="newrelic"):
        response = app.make_request("GET", "/")

    assert response.status == 200
    assert not caplog.records


def test_non_http_scope_v3(event_loop):
    async def _test():
        await simple_app_v3({"type": "cookies"}, None, None)

    with pytest.raises(ValueError):
        event_loop.run_until_complete(_test())


def test_non_http_scope_v2(event_loop):
    async def _test():
        call_me = simple_app_v2({"type": "cookies"})
        await call_me(None, None)

    with pytest.raises(ValueError):
        event_loop.run_until_complete(_test())


@pytest.mark.parametrize("app", (simple_app_v3_wrapped, simple_app_v2_wrapped, simple_app_v2_init_exc))
@validate_transaction_errors(errors=["builtins:ValueError"])
def test_exception_capture(app):
    with pytest.raises(ValueError):
        app.make_request("GET", "/exc")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import psycopg2
import psycopg2.extras
import psycopg2.extensions
import pytest

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_database_node import (
    validate_database_node,
)
from testing_support.validators.validate_transaction_slow_sql_count import (
    validate_transaction_slow_sql_count)
from newrelic.core.database_utils import SQLConnections
from testing_support.util import instance_hostname
from utils import DB_SETTINGS

from newrelic.api.background_task import background_task


_host = instance_hostname(DB_SETTINGS['host'])
_port = DB_SETTINGS['port']


class CustomConnection(psycopg2.extensions.connection):
    def __init__(self, *args, **kwargs):
        self.ready = False
        return super(CustomConnection, self).__init__(*args, **kwargs)

    def cursor(self, *args, **kwargs):
        assert self.ready  # Force a failure when generating an explain plan
        return super(CustomConnection, self).cursor(*args, **kwargs)


class CustomCursor(psycopg2.extensions.cursor):
    def __init__(self, *args, **kwargs):
        self.ready = False
        return super(CustomCursor, self).__init__(*args, **kwargs)

    def execute(self, *args, **kwargs):
        assert self.ready  # Force a failure when generating an explain plan
        return super(CustomCursor, self).execute(*args, **kwargs)


def _exercise_db(connection_factory=None, cursor_factory=None,
        cursor_kwargs=None):
    cursor_kwargs = cursor_kwargs or {}

    connect_kwargs = {'cursor_factory': cursor_factory}

    if connection_factory:
        connect_kwargs['connection_factory'] = connection_factory

    connection = psycopg2.connect(
            database=DB_SETTINGS['name'], user=DB_SETTINGS['user'],
            password=DB_SETTINGS['password'], host=DB_SETTINGS['host'],
            port=DB_SETTINGS['port'], **connect_kwargs)
    if hasattr(connection, 'ready'):
        connection.ready = True

    try:
        cursor = connection.cursor(**cursor_kwargs)
        if hasattr(cursor, 'ready'):
            cursor.ready = True

        cursor.execute("""SELECT setting from pg_settings where name=%s""",
                ('server_version',))
    finally:
        connection.close()


# Tests


def explain_plan_is_none(node):
    with SQLConnections() as connections:
        explain_plan = node.explain_plan(connections)

    assert explain_plan is None


def explain_plan_is_not_none(node):
    with SQLConnections() as connections:
        explain_plan = node.explain_plan(connections)

    assert explain_plan is not None


SCROLLABLE = (True, False)
WITHHOLD = (True, False)


@override_application_settings({
    'transaction_tracer.explain_threshold': 0.0,
    'transaction_tracer.record_sql': 'raw',
})
@validate_database_node(explain_plan_is_not_none)
@validate_transaction_slow_sql_count(1)
@background_task(name='test_explain_plan_named_cursors')
@pytest.mark.parametrize('withhold', WITHHOLD)
@pytest.mark.parametrize('scrollable', SCROLLABLE)
def test_explain_plan_named_cursors(withhold, scrollable):
    cursor_kwargs = {
        'name': 'test_explain_plan_named_cursors',
    }

    if withhold:
        cursor_kwargs['withhold'] = withhold

    if scrollable:
        cursor_kwargs['scrollable'] = scrollable

    _exercise_db(cursor_kwargs=cursor_kwargs)


# The following tests will verify that arguments are preserved for an explain
# plan by forcing a failure to be generated when explain plans are created and
# arguments are preserved
@override_application_settings({
    'transaction_tracer.explain_threshold': 0.0,
    'transaction_tracer.record_sql': 'raw',
})
@validate_database_node(explain_plan_is_none)
@validate_transaction_slow_sql_count(1)
@background_task(name='test_explain_plan_on_custom_connect_class')
def test_explain_plan_on_custom_connect_class():
    _exercise_db(connection_factory=CustomConnection)


@override_application_settings({
    'transaction_tracer.explain_threshold': 0.0,
    'transaction_tracer.record_sql': 'raw',
})
@validate_database_node(explain_plan_is_none)
@validate_transaction_slow_sql_count(1)
@background_task(name='test_explain_plan_on_custom_connect_class')
def test_explain_plan_on_custom_cursor_class_1():
    _exercise_db(cursor_factory=CustomCursor)


@override_application_settings({
    'transaction_tracer.explain_threshold': 0.0,
    'transaction_tracer.record_sql': 'raw',
})
@validate_database_node(explain_plan_is_none)
@validate_transaction_slow_sql_count(1)
@background_task(name='test_explain_plan_on_custom_connect_class')
def test_explain_plan_on_custom_cursor_class_2():
    _exercise_db(cursor_kwargs={'cursor_factory': CustomCursor})



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pytest

from newrelic.core.database_utils import SQLStatement


CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
JSON_DIR = os.path.normpath(os.path.join(CURRENT_DIR, 'fixtures',
    'sql_obfuscation'))

_parameters_list = ['obfuscated', 'dialects', 'sql', 'pathological']
_parameters = ','.join(_parameters_list)


def load_tests():
    result = []
    path = os.path.join(JSON_DIR, 'sql_obfuscation.json')
    with open(path, 'r') as fh:
        tests = json.load(fh)

    for test in tests:
        values = tuple([test.get(param, None) for param in _parameters_list])
        result.append(values)

    return result


_quoting_styles = {
    'sqlite': 'single',
    'mysql': 'single+double',
    'postgres': 'single+dollar',
    'oracle': 'single+oracle',
    'cassandra': 'single',
}


def get_quoting_styles(dialects):
    return set([_quoting_styles.get(dialect) for dialect in dialects])


class DummyDB(object):
    def __init__(self, quoting_style):
        self.quoting_style = quoting_style


@pytest.mark.parametrize(_parameters, load_tests())
def test_sql_obfuscation(obfuscated, dialects, sql, pathological):

    if pathological:
        pytest.skip()

    quoting_styles = get_quoting_styles(dialects)

    for quoting_style in quoting_styles:
        database = DummyDB(quoting_style)
        statement = SQLStatement(sql, database)
        actual_obfuscated = statement.obfuscated
        assert actual_obfuscated in obfuscated



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from newrelic.common.object_wrapper import (
    function_wrapper,
    transient_function_wrapper,
)

def validate_synthetics_event(required_attrs=None, forgone_attrs=None, should_exist=True):
    required_attrs = required_attrs or []
    forgone_attrs = forgone_attrs or []
    failed = []

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_synthetics_event(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)

        try:
            if not should_exist:
                assert instance.synthetics_events == []
            else:
                assert len(instance.synthetics_events) == 1
                event = instance.synthetics_events[0]
                assert event is not None
                assert len(event) == 3

                def _flatten(event):
                    result = {}
                    for elem in event:
                        for k, v in elem.items():
                            result[k] = v
                    return result

                flat_event = _flatten(event)

                assert "nr.guid" in flat_event, "name=%r, event=%r" % ("nr.guid", flat_event)

                for name, value in required_attrs:
                    assert name in flat_event, "name=%r, event=%r" % (name, flat_event)
                    assert flat_event[name] == value, "name=%r, value=%r, event=%r" % (name, value, flat_event)

                for name, value in forgone_attrs:
                    assert name not in flat_event, "name=%r, value=%r, event=%r" % (name, value, flat_event)
        except Exception as e:
            failed.append(e)

        return result

    @function_wrapper
    def wrapper(wrapped, instance, args, kwargs):
        _new_wrapper = _validate_synthetics_event(wrapped)
        result = _new_wrapper(*args, **kwargs)
        if failed:
            e = failed.pop()
            raise e
        return result

    return wrapper




# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import aiohttp
from testing_support.fixtures import function_not_called

version_info = tuple(int(_) for _ in aiohttp.__version__.split(".")[:2])


@function_not_called("newrelic.core.stats_engine", "StatsEngine.record_transaction")
def test_websocket(aiohttp_app):
    async def ws_write():
        ws = await aiohttp_app.client.ws_connect("/ws")
        try:
            for _ in range(2):
                result = ws.send_str("Hello")
                if hasattr(result, "__await__"):
                    await result
                msg = await ws.receive()
                assert msg.data == "/Hello"
        finally:
            await ws.close(code=1000)
            assert ws.close_code == 1000

    aiohttp_app.loop.run_until_complete(ws_write())



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pytest
import webtest

from newrelic.api.transaction import current_transaction
from newrelic.api.wsgi_application import wsgi_application
from newrelic.common.encoding_utils import DistributedTracePayload
from newrelic.common.object_wrapper import transient_function_wrapper

from testing_support.fixtures import (override_application_settings,
        validate_error_event_attributes, validate_attributes)
from testing_support.validators.validate_span_events import (
        validate_span_events)
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_event_attributes import validate_transaction_event_attributes

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
JSON_DIR = os.path.normpath(os.path.join(CURRENT_DIR, 'fixtures',
    'distributed_tracing'))

_parameters_list = ['account_id', 'comment', 'expected_metrics',
        'force_sampled_true', 'inbound_payloads', 'intrinsics',
        'major_version', 'minor_version', 'outbound_payloads',
        'raises_exception', 'span_events_enabled', 'test_name',
        'transport_type', 'trusted_account_key', 'web_transaction']
_parameters = ','.join(_parameters_list)


def load_tests():
    result = []
    path = os.path.join(JSON_DIR, 'distributed_tracing.json')
    with open(path, 'r') as fh:
        tests = json.load(fh)

    for test in tests:
        values = (test.get(param, None) for param in _parameters_list)
        param = pytest.param(*values, id=test.get('test_name'))
        result.append(param)

    return result


def override_compute_sampled(override):
    @transient_function_wrapper('newrelic.core.adaptive_sampler',
            'AdaptiveSampler.compute_sampled')
    def _override_compute_sampled(wrapped, instance, args, kwargs):
        if override:
            return True
        return wrapped(*args, **kwargs)
    return _override_compute_sampled


def assert_payload(payload, payload_assertions, major_version, minor_version):
    assert payload

    # flatten payload so it matches the test:
    #   payload['d']['ac'] -> payload['d.ac']
    d = payload.pop('d')
    for key, value in d.items():
        payload['d.%s' % key] = value

    for expected in payload_assertions.get('expected', []):
        assert expected in payload

    for unexpected in payload_assertions.get('unexpected', []):
        assert unexpected not in payload

    for key, value in payload_assertions.get('exact', {}).items():
        assert key in payload
        if isinstance(value, list):
            value = tuple(value)
        assert payload[key] == value

    assert payload['v'][0] == major_version
    assert payload['v'][1] == minor_version


@wsgi_application()
def target_wsgi_application(environ, start_response):
    status = '200 OK'
    output = b'hello world'
    response_headers = [('Content-type', 'text/html; charset=utf-8'),
                        ('Content-Length', str(len(output)))]

    txn = current_transaction()
    txn.set_transaction_name(test_settings['test_name'])

    if not test_settings['web_transaction']:
        txn.background_task = True

    if test_settings['raises_exception']:
        try:
            1 / 0
        except ZeroDivisionError:
            txn.notice_error()

    extra_inbound_payloads = test_settings['extra_inbound_payloads']
    for payload, expected_result in extra_inbound_payloads:
        result = txn.accept_distributed_trace_payload(payload,
                test_settings['transport_type'])
        assert result is expected_result

    outbound_payloads = test_settings['outbound_payloads']
    if outbound_payloads:
        for payload_assertions in outbound_payloads:
            payload = txn._create_distributed_trace_payload()
            assert_payload(payload, payload_assertions,
                    test_settings['major_version'],
                    test_settings['minor_version'])

    start_response(status, response_headers)
    return [output]


test_application = webtest.TestApp(target_wsgi_application)


@pytest.mark.parametrize(_parameters, load_tests())
def test_distributed_tracing(account_id, comment, expected_metrics,
        force_sampled_true, inbound_payloads, intrinsics, major_version,
        minor_version, outbound_payloads, raises_exception,
        span_events_enabled, test_name, transport_type, trusted_account_key,
        web_transaction):

    extra_inbound_payloads = []
    if transport_type != 'HTTP':
        # Since wsgi_application calls accept_distributed_trace_payload
        # automatically with transport_type='HTTP', we must defer this call
        # until we can specify the transport type.
        extra_inbound_payloads.append((inbound_payloads.pop(), True))
    elif not inbound_payloads:
        # In order to assert that accept_distributed_trace_payload returns
        # False in this instance, we defer.
        extra_inbound_payloads.append((inbound_payloads, False))
    elif len(inbound_payloads) > 1:
        for payload in inbound_payloads[1:]:
            extra_inbound_payloads.append((payload, False))

    global test_settings
    test_settings = {
        'test_name': test_name,
        'web_transaction': web_transaction,
        'raises_exception': raises_exception,
        'extra_inbound_payloads': extra_inbound_payloads,
        'outbound_payloads': outbound_payloads,
        'transport_type': transport_type,
        'major_version': major_version,
        'minor_version': minor_version,
    }

    override_settings = {
        'distributed_tracing.enabled': True,
        'span_events.enabled': span_events_enabled,
        'account_id': account_id,
        'trusted_account_key': trusted_account_key,
    }

    common_required = intrinsics['common']['expected']
    common_forgone = intrinsics['common']['unexpected']
    common_exact = intrinsics['common'].get('exact', {})

    txn_intrinsics = intrinsics.get('Transaction', {})
    txn_event_required = {'agent': [], 'user': [],
            'intrinsic': txn_intrinsics.get('expected', [])}
    txn_event_required['intrinsic'].extend(common_required)
    txn_event_forgone = {'agent': [], 'user': [],
            'intrinsic': txn_intrinsics.get('unexpected', [])}
    txn_event_forgone['intrinsic'].extend(common_forgone)
    txn_event_exact = {'agent': {}, 'user': {},
            'intrinsic': txn_intrinsics.get('exact', {})}
    txn_event_exact['intrinsic'].update(common_exact)

    headers = {}
    if inbound_payloads:
        payload = DistributedTracePayload(inbound_payloads[0])
        headers['newrelic'] = payload.http_safe()

    @validate_transaction_metrics(test_name,
            rollup_metrics=expected_metrics,
            background_task=not web_transaction)
    @validate_transaction_event_attributes(
            txn_event_required, txn_event_forgone, txn_event_exact)
    @validate_attributes('intrinsic', common_required, common_forgone)
    @override_compute_sampled(force_sampled_true)
    @override_application_settings(override_settings)
    def _test():
        response = test_application.get('/', headers=headers)
        assert 'X-NewRelic-App-Data' not in response.headers

    if 'Span' in intrinsics:
        span_intrinsics = intrinsics.get('Span')
        span_expected = span_intrinsics.get('expected', [])
        span_expected.extend(common_required)
        span_unexpected = span_intrinsics.get('unexpected', [])
        span_unexpected.extend(common_forgone)
        span_exact = span_intrinsics.get('exact', {})
        span_exact.update(common_exact)

        _test = validate_span_events(exact_intrinsics=span_exact,
            expected_intrinsics=span_expected,
            unexpected_intrinsics=span_unexpected)(_test)
    elif not span_events_enabled:
        _test = validate_span_events(count=0)(_test)

    if raises_exception:
        error_event_required = {'agent': [], 'user': [],
                'intrinsic': common_required}
        error_event_forgone = {'agent': [], 'user': [],
                'intrinsic': common_forgone}
        error_event_exact = {'agent': {}, 'user': {},
                'intrinsic': common_exact}
        _test = validate_error_event_attributes(error_event_required,
                error_event_forgone, error_event_exact)(_test)

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.object_wrapper import (transient_function_wrapper,
        function_wrapper)


def validate_metric_payload(metrics=[]):
    # Validates metrics as they are sent to the collector. Useful for testing
    # Supportability metrics that are created at harvest time.

    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):

        recorded_metrics = []

        @transient_function_wrapper('newrelic.core.agent_protocol',
                'AgentProtocol.send')
        def send_request_wrapper(wrapped, instance, args, kwargs):
            def _bind_params(method, payload=(), *args, **kwargs):
                return method, payload

            method, payload = _bind_params(*args, **kwargs)

            if method == 'metric_data' and payload:
                sent_metrics = {}
                for metric_info, metric_values in payload[3]:
                    metric_key = (metric_info['name'], metric_info['scope'])
                    sent_metrics[metric_key] = metric_values

                recorded_metrics.append(sent_metrics)

            return wrapped(*args, **kwargs)

        wrapped = send_request_wrapper(wrapped)
        val = wrapped(*args, **kwargs)
        assert recorded_metrics

        for sent_metrics in recorded_metrics:
            for metric, count in metrics:
                # only look for unscoped metrics
                unscoped_metric = (metric, '')
                if not count:
                    assert unscoped_metric not in sent_metrics, unscoped_metric
                else:
                    assert unscoped_metric in sent_metrics, unscoped_metric
                    metric_values = sent_metrics[unscoped_metric]
                    assert metric_values[0] == count, "%s: Expected: %d Got: %d" % (metric, count, metric_values[0])

        return val

    return _validate_wrapper



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pymysql

from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_database_trace_inputs import validate_database_trace_inputs

from testing_support.db_settings import mysql_settings

from newrelic.api.background_task import background_task

DB_SETTINGS = mysql_settings()[0]
TABLE_NAME = "datastore_pymysql_" + DB_SETTINGS["namespace"]
PROCEDURE_NAME = "hello_" + DB_SETTINGS["namespace"]


def execute_db_calls_with_cursor(cursor):
    cursor.execute("""drop table if exists %s""" % TABLE_NAME)

    cursor.execute("""create table %s """ % TABLE_NAME +
           """(a integer, b real, c text)""")

    cursor.executemany("""insert into %s """ % TABLE_NAME +
            """values (%s, %s, %s)""", [(1, 1.0, '1.0'),
            (2, 2.2, '2.2'), (3, 3.3, '3.3')])

    cursor.execute("""select * from %s""" % TABLE_NAME)

    for row in cursor: pass

    cursor.execute("""update %s""" % TABLE_NAME + """ set a=%s, b=%s, """
            """c=%s where a=%s""", (4, 4.0, '4.0', 1))

    cursor.execute("""delete from %s where a=2""" % TABLE_NAME)
    cursor.execute("""drop procedure if exists %s""" % PROCEDURE_NAME)
    cursor.execute("""CREATE PROCEDURE %s()
                      BEGIN
                        SELECT 'Hello World!';
                      END""" % PROCEDURE_NAME)

    cursor.callproc(PROCEDURE_NAME)


_test_execute_via_cursor_scoped_metrics = [
        ('Function/pymysql:Connect', 1),
        ('Datastore/statement/MySQL/%s/select' % TABLE_NAME, 1),
        ('Datastore/statement/MySQL/%s/insert' % TABLE_NAME, 1),
        ('Datastore/statement/MySQL/%s/update' % TABLE_NAME, 1),
        ('Datastore/statement/MySQL/%s/delete' % TABLE_NAME, 1),
        ('Datastore/operation/MySQL/drop', 2),
        ('Datastore/operation/MySQL/create', 2),
        ('Datastore/statement/MySQL/%s/call' % PROCEDURE_NAME, 1),
        ('Datastore/operation/MySQL/commit', 2),
        ('Datastore/operation/MySQL/rollback', 1)]

_test_execute_via_cursor_rollup_metrics = [
        ('Datastore/all', 13),
        ('Datastore/allOther', 13),
        ('Datastore/MySQL/all', 13),
        ('Datastore/MySQL/allOther', 13),
        ('Datastore/statement/MySQL/%s/select' % TABLE_NAME, 1),
        ('Datastore/statement/MySQL/%s/insert' % TABLE_NAME, 1),
        ('Datastore/statement/MySQL/%s/update' % TABLE_NAME, 1),
        ('Datastore/statement/MySQL/%s/delete' % TABLE_NAME, 1),
        ('Datastore/operation/MySQL/select', 1),
        ('Datastore/operation/MySQL/insert', 1),
        ('Datastore/operation/MySQL/update', 1),
        ('Datastore/operation/MySQL/delete', 1),
        ('Datastore/statement/MySQL/%s/call' % PROCEDURE_NAME, 1),
        ('Datastore/operation/MySQL/call', 1),
        ('Datastore/operation/MySQL/drop', 2),
        ('Datastore/operation/MySQL/create', 2),
        ('Datastore/operation/MySQL/commit', 2),
        ('Datastore/operation/MySQL/rollback', 1)]

@validate_transaction_metrics('test_database:test_execute_via_cursor',
        scoped_metrics=_test_execute_via_cursor_scoped_metrics,
        rollup_metrics=_test_execute_via_cursor_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_execute_via_cursor():
    connection = pymysql.connect(db=DB_SETTINGS['name'],
            user=DB_SETTINGS['user'], passwd=DB_SETTINGS['password'],
            host=DB_SETTINGS['host'], port=DB_SETTINGS['port'])

    with connection.cursor() as cursor:
        execute_db_calls_with_cursor(cursor)

    connection.commit()
    connection.rollback()
    connection.commit()


_test_execute_via_cursor_context_mangaer_scoped_metrics = [
        ('Function/pymysql:Connect', 1),
        ('Datastore/statement/MySQL/%s/select' % TABLE_NAME, 1),
        ('Datastore/statement/MySQL/%s/insert' % TABLE_NAME, 1),
        ('Datastore/statement/MySQL/%s/update' % TABLE_NAME, 1),
        ('Datastore/statement/MySQL/%s/delete' % TABLE_NAME, 1),
        ('Datastore/operation/MySQL/drop', 2),
        ('Datastore/operation/MySQL/create', 2),
        ('Datastore/statement/MySQL/%s/call' % PROCEDURE_NAME, 1),
        ('Datastore/operation/MySQL/commit', 2),
        ('Datastore/operation/MySQL/rollback', 1)]

_test_execute_via_cursor_context_mangaer_rollup_metrics = [
        ('Datastore/all', 13),
        ('Datastore/allOther', 13),
        ('Datastore/MySQL/all', 13),
        ('Datastore/MySQL/allOther', 13),
        ('Datastore/statement/MySQL/%s/select' % TABLE_NAME, 1),
        ('Datastore/statement/MySQL/%s/insert' % TABLE_NAME, 1),
        ('Datastore/statement/MySQL/%s/update' % TABLE_NAME, 1),
        ('Datastore/statement/MySQL/%s/delete' % TABLE_NAME, 1),
        ('Datastore/operation/MySQL/select', 1),
        ('Datastore/operation/MySQL/insert', 1),
        ('Datastore/operation/MySQL/update', 1),
        ('Datastore/operation/MySQL/delete', 1),
        ('Datastore/statement/MySQL/%s/call' % PROCEDURE_NAME, 1),
        ('Datastore/operation/MySQL/call', 1),
        ('Datastore/operation/MySQL/drop', 2),
        ('Datastore/operation/MySQL/create', 2),
        ('Datastore/operation/MySQL/commit', 2),
        ('Datastore/operation/MySQL/rollback', 1)]


@validate_transaction_metrics(
        'test_database:test_execute_via_cursor_context_manager',
        scoped_metrics=_test_execute_via_cursor_context_mangaer_scoped_metrics,
        rollup_metrics=_test_execute_via_cursor_context_mangaer_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_execute_via_cursor_context_manager():
    connection = pymysql.connect(db=DB_SETTINGS['name'],
            user=DB_SETTINGS['user'], passwd=DB_SETTINGS['password'],
            host=DB_SETTINGS['host'], port=DB_SETTINGS['port'])
    cursor = connection.cursor()

    with cursor:
        execute_db_calls_with_cursor(cursor)

    connection.commit()
    connection.rollback()
    connection.commit()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from collections import namedtuple

from newrelic.core.attribute_filter import (
    DST_ALL,
    DST_ERROR_COLLECTOR,
    DST_SPAN_EVENTS,
    DST_TRANSACTION_EVENTS,
    DST_TRANSACTION_SEGMENTS,
    DST_TRANSACTION_TRACER,
)
from newrelic.packages import six

_logger = logging.getLogger(__name__)

_Attribute = namedtuple("_Attribute", ["name", "value", "destinations"])

# The following destinations are created here, never changed, and only
# used in create_agent_attributes. It is placed at the module level here
# as an optimization.

# All agent attributes go to transaction traces and error traces by default.

_DESTINATIONS = DST_ERROR_COLLECTOR | DST_TRANSACTION_TRACER | DST_TRANSACTION_SEGMENTS
_DESTINATIONS_WITH_EVENTS = _DESTINATIONS | DST_TRANSACTION_EVENTS | DST_SPAN_EVENTS

# The following subset goes to transaction events by default.

_TRANSACTION_EVENT_DEFAULT_ATTRIBUTES = set(
    (
        "host.displayName",
        "request.method",
        "request.headers.contentType",
        "request.headers.contentLength",
        "request.uri",
        "response.status",
        "request.headers.accept",
        "response.headers.contentLength",
        "response.headers.contentType",
        "request.headers.host",
        "request.headers.userAgent",
        "message.queueName",
        "message.routingKey",
        "http.url",
        "http.statusCode",
        "aws.requestId",
        "aws.operation",
        "aws.lambda.arn",
        "aws.lambda.coldStart",
        "aws.lambda.eventSource.arn",
        "db.collection",
        "db.instance",
        "db.operation",
        "db.statement",
        "error.class",
        "error.message",
        "error.expected",
        "peer.hostname",
        "peer.address",
        "graphql.field.name",
        "graphql.field.parentType",
        "graphql.field.path",
        "graphql.field.returnType",
        "graphql.operation.name",
        "graphql.operation.type",
        "graphql.operation.query",
        "code.filepath",
        "code.function",
        "code.lineno",
        "code.namespace",
    )
)

MAX_NUM_USER_ATTRIBUTES = 128
MAX_ATTRIBUTE_LENGTH = 255
MAX_64_BIT_INT = 2**63 - 1
MAX_LOG_MESSAGE_LENGTH = 32768


class NameTooLongException(Exception):
    pass


class NameIsNotStringException(Exception):
    pass


class IntTooLargeException(Exception):
    pass


class CastingFailureException(Exception):
    pass


class Attribute(_Attribute):
    def __repr__(self):
        return "Attribute(name=%r, value=%r, destinations=%r)" % (self.name, self.value, bin(self.destinations))


def create_attributes(attr_dict, destinations, attribute_filter):
    attributes = []

    for k, v in attr_dict.items():
        dest = attribute_filter.apply(k, destinations)
        attributes.append(Attribute(k, v, dest))

    return attributes


def create_agent_attributes(attr_dict, attribute_filter):
    attributes = []

    for k, v in attr_dict.items():
        if v is None:
            continue

        if k in _TRANSACTION_EVENT_DEFAULT_ATTRIBUTES:
            dest = attribute_filter.apply(k, _DESTINATIONS_WITH_EVENTS)
        else:
            dest = attribute_filter.apply(k, _DESTINATIONS)

        attributes.append(Attribute(k, v, dest))

    return attributes


def resolve_user_attributes(attr_dict, attribute_filter, target_destination, attr_class=dict):
    u_attrs = attr_class()

    for attr_name, attr_value in attr_dict.items():
        if attr_value is None:
            continue

        dest = attribute_filter.apply(attr_name, DST_ALL)

        if dest & target_destination:
            u_attrs[attr_name] = attr_value

    return u_attrs


def resolve_agent_attributes(attr_dict, attribute_filter, target_destination, attr_class=dict):
    a_attrs = attr_class()

    for attr_name, attr_value in attr_dict.items():
        if attr_value is None:
            continue

        if attr_name in _TRANSACTION_EVENT_DEFAULT_ATTRIBUTES:
            dest = attribute_filter.apply(attr_name, _DESTINATIONS_WITH_EVENTS)
        else:
            dest = attribute_filter.apply(attr_name, _DESTINATIONS)

        if dest & target_destination:
            a_attrs[attr_name] = attr_value

    return a_attrs


def create_user_attributes(attr_dict, attribute_filter):
    destinations = DST_ALL
    return create_attributes(attr_dict, destinations, attribute_filter)


def truncate(text, maxsize=MAX_ATTRIBUTE_LENGTH, encoding="utf-8", ending=None):

    # Truncate text so that its byte representation
    # is no longer than maxsize bytes.

    # If text is unicode (Python 2 or 3), return unicode.
    # If text is a Python 2 string, return str.

    if isinstance(text, six.text_type):
        truncated = _truncate_unicode(text, maxsize, encoding)
    else:
        truncated = _truncate_bytes(text, maxsize)
        ending = ending and ending.encode(encoding)

    if ending and truncated != text:
        truncated = truncated[: -len(ending)] + ending

    return truncated


def _truncate_unicode(u, maxsize, encoding="utf-8"):
    encoded = u.encode(encoding)[:maxsize]
    return encoded.decode(encoding, "ignore")


def _truncate_bytes(s, maxsize):
    return s[:maxsize]


def check_name_length(name, max_length=MAX_ATTRIBUTE_LENGTH, encoding="utf-8"):
    trunc_name = truncate(name, max_length, encoding)
    if name != trunc_name:
        raise NameTooLongException()


def check_name_is_string(name):
    if not isinstance(name, (six.text_type, six.binary_type)):
        raise NameIsNotStringException()


def check_max_int(value, max_int=MAX_64_BIT_INT):
    if isinstance(value, six.integer_types) and value > max_int:
        raise IntTooLargeException()


def process_user_attribute(name, value, max_length=MAX_ATTRIBUTE_LENGTH, ending=None):

    # Perform all necessary checks on a potential attribute.
    #
    # Returns:
    #       (name, value) if attribute is OK.
    #       (NONE, NONE) if attribute isn't.
    #
    # If any of these checks fail, they will raise an exception, so we
    # log a message, and return (None, None).

    FAILED_RESULT = (None, None)

    try:
        check_name_is_string(name)
        check_name_length(name)
        check_max_int(value)

        value = sanitize(value)

    except NameIsNotStringException:
        _logger.debug("Attribute name must be a string. Dropping " "attribute: %r=%r", name, value)
        return FAILED_RESULT

    except NameTooLongException:
        _logger.debug("Attribute name exceeds maximum length. Dropping " "attribute: %r=%r", name, value)
        return FAILED_RESULT

    except IntTooLargeException:
        _logger.debug("Attribute value exceeds maximum integer value. " "Dropping attribute: %r=%r", name, value)
        return FAILED_RESULT

    except CastingFailureException:
        _logger.debug("Attribute value cannot be cast to a string. " "Dropping attribute: %r=%r", name, value)
        return FAILED_RESULT

    else:

        # Check length after casting

        valid_types_text = (six.text_type, six.binary_type)

        if isinstance(value, valid_types_text):
            trunc_value = truncate(value, maxsize=max_length, ending=ending)
            if value != trunc_value:
                _logger.debug(
                    "Attribute value exceeds maximum length " "(%r bytes). Truncating value: %r=%r.",
                    max_length,
                    name,
                    trunc_value,
                )

            value = trunc_value

        return (name, value)


def sanitize(value):

    # Return value unchanged, if it's a valid type that is supported by
    # Insights. Otherwise, convert value to a string.
    #
    # Raise CastingFailureException, if str(value) somehow fails.

    valid_value_types = (six.text_type, six.binary_type, bool, float, six.integer_types)

    if not isinstance(value, valid_value_types):
        original = value

        try:
            value = str(value)
        except Exception:
            raise CastingFailureException()
        else:
            _logger.debug(
                "Attribute value is of type: %r. Casting %r to " "string: %s", type(original), original, value
            )

    return value



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pytest
import webtest
from newrelic.packages import six

from newrelic.api.transaction import current_transaction
from newrelic.api.wsgi_application import wsgi_application
from newrelic.common.object_wrapper import transient_function_wrapper
from testing_support.validators.validate_span_events import (
        validate_span_events)
from testing_support.fixtures import (override_application_settings,
        validate_attributes)
from newrelic.common.encoding_utils import W3CTraceState
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_event_attributes import validate_transaction_event_attributes

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
JSON_DIR = os.path.normpath(os.path.join(CURRENT_DIR, 'fixtures',
    'distributed_tracing'))

_parameters_list = ('test_name', 'trusted_account_key', 'account_id',
        'web_transaction', 'raises_exception', 'force_sampled_true',
        'span_events_enabled', 'transport_type', 'inbound_headers',
        'outbound_payloads', 'intrinsics', 'expected_metrics')

_parameters = ','.join(_parameters_list)


XFAIL_TESTS = [
    'spans_disabled_root',
    'missing_traceparent',
    'missing_traceparent_and_tracestate',
    'w3c_and_newrelc_headers_present_error_parsing_traceparent'
]

def load_tests():
    result = []
    path = os.path.join(JSON_DIR, 'trace_context.json')
    with open(path, 'r') as fh:
        tests = json.load(fh)

    for test in tests:
        values = (test.get(param, None) for param in _parameters_list)
        param = pytest.param(*values, id=test.get('test_name'))
        result.append(param)

    return result


ATTR_MAP = {
    'traceparent.version': 0,
    'traceparent.trace_id': 1,
    'traceparent.parent_id': 2,
    'traceparent.trace_flags': 3,
    'tracestate.version': 0,
    'tracestate.parent_type': 1,
    'tracestate.parent_account_id': 2,
    'tracestate.parent_application_id': 3,
    'tracestate.span_id': 4,
    'tracestate.transaction_id': 5,
    'tracestate.sampled': 6,
    'tracestate.priority': 7,
    'tracestate.timestamp': 8,
    'tracestate.tenant_id': None,
}


def validate_outbound_payload(actual, expected, trusted_account_key):
    traceparent = ''
    tracestate = ''
    for key, value in actual:
        if key == 'traceparent':
            traceparent = value.split('-')
        elif key == 'tracestate':
            vendors = W3CTraceState.decode(value)
            nr_entry = vendors.pop(trusted_account_key + '@nr', '')
            tracestate = nr_entry.split('-')
    exact_values = expected.get('exact', {})
    expected_attrs = expected.get('expected', [])
    unexpected_attrs = expected.get('unexpected', [])
    expected_vendors = expected.get('vendors', [])
    for key, value in exact_values.items():
        header = traceparent if key.startswith('traceparent.') else tracestate
        attr = ATTR_MAP[key]
        if attr is not None:
            if isinstance(value, bool):
                assert header[attr] == str(int(value))
            elif isinstance(value, int):
                assert int(header[attr]) == value
            else:
                assert header[attr] == str(value)

    for key in expected_attrs:
        header = traceparent if key.startswith('traceparent.') else tracestate
        attr = ATTR_MAP[key]
        if attr is not None:
            assert header[attr], key

    for key in unexpected_attrs:
        header = traceparent if key.startswith('traceparent.') else tracestate
        attr = ATTR_MAP[key]
        if attr is not None:
            assert not header[attr], key

    for vendor in expected_vendors:
        assert vendor in vendors


@wsgi_application()
def target_wsgi_application(environ, start_response):
    transaction = current_transaction()

    if not environ['.web_transaction']:
        transaction.background_task = True

    if environ['.raises_exception']:
        try:
            raise ValueError("oops")
        except:
            transaction.notice_error()

    if '.inbound_headers' in environ:
        transaction.accept_distributed_trace_headers(
            environ['.inbound_headers'],
            transport_type=environ['.transport_type'],
        )

    payloads = []
    for _ in range(environ['.outbound_calls']):
        payloads.append([])
        transaction.insert_distributed_trace_headers(payloads[-1])

    start_response('200 OK', [('Content-Type', 'application/json')])
    return [json.dumps(payloads).encode('utf-8')]


test_application = webtest.TestApp(target_wsgi_application)


def override_compute_sampled(override):
    @transient_function_wrapper('newrelic.core.adaptive_sampler',
            'AdaptiveSampler.compute_sampled')
    def _override_compute_sampled(wrapped, instance, args, kwargs):
        if override:
            return True
        return wrapped(*args, **kwargs)
    return _override_compute_sampled


@pytest.mark.parametrize(_parameters, load_tests())
def test_trace_context(test_name, trusted_account_key, account_id,
        web_transaction, raises_exception, force_sampled_true,
        span_events_enabled, transport_type, inbound_headers,
        outbound_payloads, intrinsics, expected_metrics):

    if test_name in XFAIL_TESTS:
        pytest.xfail("Waiting on cross agent tests update.")
    # Prepare assertions
    if not intrinsics:
        intrinsics = {}

    common = intrinsics.get('common', {})
    common_required = common.get('expected', [])
    common_forgone = common.get('unexpected', [])
    common_exact = common.get('exact', {})

    txn_intrinsics = intrinsics.get('Transaction', {})
    txn_event_required = {'agent': [], 'user': [],
            'intrinsic': txn_intrinsics.get('expected', [])}
    txn_event_required['intrinsic'].extend(common_required)
    txn_event_forgone = {'agent': [], 'user': [],
            'intrinsic': txn_intrinsics.get('unexpected', [])}
    txn_event_forgone['intrinsic'].extend(common_forgone)
    txn_event_exact = {'agent': {}, 'user': {},
            'intrinsic': txn_intrinsics.get('exact', {})}
    txn_event_exact['intrinsic'].update(common_exact)

    override_settings = {
        'distributed_tracing.enabled': True,
        'span_events.enabled': span_events_enabled,
        'account_id': account_id,
        'trusted_account_key': trusted_account_key,
    }

    extra_environ = {
        '.web_transaction': web_transaction,
        '.raises_exception': raises_exception,
        '.transport_type': transport_type,
        '.outbound_calls': outbound_payloads and len(outbound_payloads) or 0,
    }

    inbound_headers = inbound_headers and inbound_headers[0] or None
    if transport_type != 'HTTP':
        extra_environ['.inbound_headers'] = inbound_headers
        inbound_headers = None
    elif six.PY2 and inbound_headers:
        inbound_headers = {
                k.encode('utf-8'): v.encode('utf-8')
                for k, v in inbound_headers.items()}

    @validate_transaction_metrics(test_name,
            group="Uri",
            rollup_metrics=expected_metrics,
            background_task=not web_transaction)
    @validate_transaction_event_attributes(
            txn_event_required, txn_event_forgone, txn_event_exact)
    @validate_attributes('intrinsic', common_required, common_forgone)
    @override_application_settings(override_settings)
    @override_compute_sampled(force_sampled_true)
    def _test():
        return test_application.get(
            '/' + test_name,
            headers=inbound_headers,
            extra_environ=extra_environ,
        )

    if 'Span' in intrinsics:
        span_intrinsics = intrinsics.get('Span')
        span_expected = span_intrinsics.get('expected', [])
        span_expected.extend(common_required)
        span_unexpected = span_intrinsics.get('unexpected', [])
        span_unexpected.extend(common_forgone)
        span_exact = span_intrinsics.get('exact', {})
        span_exact.update(common_exact)

        _test = validate_span_events(exact_intrinsics=span_exact,
            expected_intrinsics=span_expected,
            unexpected_intrinsics=span_unexpected)(_test)
    elif not span_events_enabled:
        _test = validate_span_events(count=0)(_test)

    response = _test()
    assert response.status == '200 OK'
    payloads = response.json
    if outbound_payloads:
        assert len(payloads) == len(outbound_payloads)
        for actual, expected in zip(payloads, outbound_payloads):
            validate_outbound_payload(actual, expected, trusted_account_key)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webtest

from bottle import __version__ as version
from bottle import route, error, default_app, HTTPError

version = [int(x) for x in version.split('-')[0].split('.')]

if len(version) == 2:
    version.append(0)

version = tuple(version)

@route('/index')
def index_page():
    return 'INDEX RESPONSE'

@route('/error')
def error_page():
    raise RuntimeError('RUNTIME ERROR')

@error(404)
def error404_page(error):
    return 'NOT FOUND'

@route('/html_insertion')
def index_page():
    return ('<!DOCTYPE html><html><head>Some header</head>'
            '<body><h1>My First Heading</h1><p>My first paragraph.</p>'
            '</body></html>')

if version >= (0, 9, 0):
    from bottle import auth_basic

    def auth_check(username, password):
        return username == 'user' and password == 'password'

    @route('/auth')
    @auth_basic(auth_check)
    def auth_basic_page():
        return 'AUTH OKAY'

    def plugin_error(callback):
        def wrapper(*args, **kwargs):
            raise HTTPError(403, 'Forbidden')
        return wrapper

    @route('/plugin_error', apply=[plugin_error], skip=['json'])
    def plugin_error_page():
        return 'PLUGIN RESPONSE'

application = default_app()
target_application = webtest.TestApp(application)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import string
import pytest
import memcache

from testing_support.fixtures import (code_coverage_fixture,
        collector_agent_registration_fixture, collector_available_fixture)

from testing_support.db_settings import memcached_settings

_coverage_source = [
    'newrelic.api.memcache_trace',
    'newrelic.hooks.datastore_memcache',
]

code_coverage = code_coverage_fixture(source=_coverage_source)

_default_settings = {
    'transaction_tracer.explain_threshold': 0.0,
    'transaction_tracer.transaction_threshold': 0.0,
    'transaction_tracer.stack_trace_threshold': 0.0,
    'debug.log_data_collector_payloads': True,
    'debug.record_transaction_failure': True
}

collector_agent_registration = collector_agent_registration_fixture(
        app_name='Python Agent Test (datastore_memcache)',
        default_settings=_default_settings,
        linked_applications=['Python Agent Test (datastore)'])

@pytest.fixture(scope='session')
def memcached_multi():
    """Generate keys that will go onto different servers"""
    DB_SETTINGS = memcached_settings()
    db_servers = ['%s:%s' % (s['host'], s['port']) for s in DB_SETTINGS]

    clients = [memcache.Client([s]) for s in db_servers]
    client_all = memcache.Client(db_servers)
    num_servers = len(db_servers)

    for try_num in range(10 * num_servers):
        multi_dict = {}
        for i in range(num_servers):
            random_chars = (random.choice(string.ascii_uppercase)
                    for _ in range(10))
            key_candidate = ''.join(random_chars)
            multi_dict[key_candidate] = key_candidate

        client_all.set_multi(multi_dict)

        server_hit = [False] * num_servers

        for key in multi_dict.keys():
            for i in range(num_servers):
                if clients[i].get(key):
                    server_hit[i] = True

        if all(server_hit):
            break
    else:
        assert False, "memcached_multi failed to map keys to multiple servers."

    return multi_dict



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testing_support.fixtures import core_application_stats_engine

from newrelic.common.object_wrapper import function_wrapper


def validate_application_error_event_count(num_errors):
    """Validate error event data for a single error occurring outside of a
    transaction.
    """

    @function_wrapper
    def _validate_application_error_event_count(wrapped, instace, args, kwargs):

        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:

            stats = core_application_stats_engine(None)
            assert len(list(stats.error_events)) == num_errors

        return result

    return _validate_application_error_event_count



#!/usr/bin/env python3.8
import fileinput
import os

GROUP_NUMBER = int(os.environ["GROUP_NUMBER"]) - 1
TOTAL_GROUPS = int(os.environ["TOTAL_GROUPS"])


def main(f):
    environments = [e.rstrip() for e in f]
    filtered_envs = environments[GROUP_NUMBER::TOTAL_GROUPS]
    print(",".join(filtered_envs))


if __name__ == "__main__":
    with fileinput.input() as f:
        main(f)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import webtest
from testing_support.fixtures import (
    override_application_settings,
    validate_custom_event_collector_json,
)
from testing_support.sample_applications import (
    simple_app,
    simple_custom_event_app,
    simple_exceptional_app,
)
from testing_support.validators.validate_error_event_collector_json import (
    validate_error_event_collector_json,
)
from testing_support.validators.validate_error_trace_collector_json import (
    validate_error_trace_collector_json,
)
from testing_support.validators.validate_log_event_collector_json import (
    validate_log_event_collector_json,
)
from testing_support.validators.validate_transaction_event_collector_json import (
    validate_transaction_event_collector_json,
)
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

exceptional_application = webtest.TestApp(simple_exceptional_app)
normal_application = webtest.TestApp(simple_app)
custom_event_application = webtest.TestApp(simple_custom_event_app)


@validate_error_trace_collector_json()
def test_error_trace_json():
    try:
        exceptional_application.get("/")
    except ValueError:
        pass


@validate_error_event_collector_json()
def test_error_event_json():
    try:
        exceptional_application.get("/")
    except ValueError:
        pass


@validate_tt_collector_json()
def test_transaction_trace_json():
    normal_application.get("/")


@validate_tt_collector_json(exclude_request_uri=True)
@override_application_settings({"attributes.exclude": set(("request.uri",))})
def test_transaction_trace_json_no_request_uri():
    normal_application.get("/")


@validate_transaction_event_collector_json()
def test_transaction_event_json():
    normal_application.get("/")


@validate_custom_event_collector_json()
def test_custom_event_json():
    custom_event_application.get("/")


@pytest.mark.xfail(reason="Unwritten validator")
@validate_log_event_collector_json
def test_log_event_json():
    normal_application.get("/")
    raise NotImplementedError("Fix my validator")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import boto3
import moto
import pytest

from newrelic.api.background_task import background_task
from testing_support.fixtures import (
        validate_tt_segment_params, override_application_settings)
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics

MOTO_VERSION = tuple(int(v) for v in moto.__version__.split('.')[:3])

# patch earlier versions of moto to support py37
if sys.version_info >= (3, 7) and MOTO_VERSION <= (1, 3, 1):
    import re
    moto.packages.responses.responses.re._pattern_type = re.Pattern

AWS_ACCESS_KEY_ID = 'AAAAAAAAAAAACCESSKEY'
AWS_SECRET_ACCESS_KEY = 'AAAAAASECRETKEY'
AWS_REGION_NAME = 'us-east-1'
SNS_URL = 'sns-us-east-1.amazonaws.com'
TOPIC = 'arn:aws:sns:us-east-1:123456789012:some-topic'
sns_metrics = [
        ('MessageBroker/SNS/Topic'
        '/Produce/Named/%s' % TOPIC, 1)]
sns_metrics_phone = [
        ('MessageBroker/SNS/Topic'
        '/Produce/Named/PhoneNumber', 1)]


@override_application_settings({'distributed_tracing.enabled': True})
@validate_span_events(expected_agents=('aws.requestId',), count=2)
@validate_span_events(exact_agents={'aws.operation': 'CreateTopic'}, count=1)
@validate_span_events(exact_agents={'aws.operation': 'Publish'}, count=1)
@validate_tt_segment_params(present_params=('aws.requestId',))
@pytest.mark.parametrize('topic_argument', ('TopicArn', 'TargetArn'))
@validate_transaction_metrics('test_boto3_sns:test_publish_to_sns_topic',
        scoped_metrics=sns_metrics, rollup_metrics=sns_metrics,
        background_task=True)
@background_task()
@moto.mock_sns
def test_publish_to_sns_topic(topic_argument):
    conn = boto3.client('sns',
            aws_access_key_id=AWS_ACCESS_KEY_ID,
            aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
            region_name=AWS_REGION_NAME)

    topic_arn = conn.create_topic(Name='some-topic')['TopicArn']

    kwargs = {topic_argument: topic_arn}
    published_message = conn.publish(Message='my msg', **kwargs)
    assert 'MessageId' in published_message


@override_application_settings({'distributed_tracing.enabled': True})
@validate_span_events(expected_agents=('aws.requestId',), count=3)
@validate_span_events(exact_agents={'aws.operation': 'CreateTopic'}, count=1)
@validate_span_events(exact_agents={'aws.operation': 'Subscribe'}, count=1)
@validate_span_events(exact_agents={'aws.operation': 'Publish'}, count=1)
@validate_tt_segment_params(present_params=('aws.requestId',))
@validate_transaction_metrics('test_boto3_sns:test_publish_to_sns_phone',
        scoped_metrics=sns_metrics_phone, rollup_metrics=sns_metrics_phone,
        background_task=True)
@background_task()
@moto.mock_sns
def test_publish_to_sns_phone():
    conn = boto3.client('sns',
            aws_access_key_id=AWS_ACCESS_KEY_ID,
            aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
            region_name=AWS_REGION_NAME)

    topic_arn = conn.create_topic(Name='some-topic')['TopicArn']
    conn.subscribe(TopicArn=topic_arn, Protocol='sms', Endpoint='5555555555')

    published_message = conn.publish(
            PhoneNumber='5555555555', Message='my msg')
    assert 'MessageId' in published_message



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import webtest

from newrelic.packages import six
from newrelic.core.config import global_settings

from testing_support.fixtures import (
    override_generic_settings,
    function_not_called)
from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics
import django

DJANGO_VERSION = tuple(map(int, django.get_version().split('.')[:2]))


@pytest.fixture(scope='module')
def target_application():
    from wsgi import application
    test_application = webtest.TestApp(application)
    return test_application


if DJANGO_VERSION >= (1, 10):
    url_module_path = 'django.urls.resolvers'

    # Django 1.10 new style middleware removed individual process_* methods.
    # All middleware in Django 1.10+ is called through the __call__ methods on
    # middlwares.
    process_request_method = ''
    process_view_method = ''
    process_response_method = ''
else:
    url_module_path = 'django.core.urlresolvers'
    process_request_method = '.process_request'
    process_view_method = '.process_view'
    process_response_method = '.process_response'

if DJANGO_VERSION >= (2, 0):
    url_resolver_cls = 'URLResolver'
else:
    url_resolver_cls = 'RegexURLResolver'

_scoped_metrics = [
        ('Function/django.core.handlers.wsgi:WSGIHandler.__call__', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        (('Function/django.middleware.common:'
                'CommonMiddleware' + process_request_method), 1),
        (('Function/django.contrib.sessions.middleware:'
                'SessionMiddleware' + process_request_method), 1),
        (('Function/django.contrib.auth.middleware:'
                'AuthenticationMiddleware' + process_request_method), 1),
        (('Function/django.contrib.messages.middleware:'
                'MessageMiddleware' + process_request_method), 1),
        (('Function/%s:' % url_module_path +
                '%s.resolve' % url_resolver_cls), 1),
        (('Function/django.middleware.csrf:'
                'CsrfViewMiddleware' + process_view_method), 1),
        (('Function/django.contrib.messages.middleware:'
                'MessageMiddleware' + process_response_method), 1),
        (('Function/django.middleware.csrf:'
                'CsrfViewMiddleware' + process_response_method), 1),
        (('Function/django.contrib.sessions.middleware:'
                'SessionMiddleware' + process_response_method), 1),
        (('Function/django.middleware.common:'
                'CommonMiddleware' + process_response_method), 1),
]

_test_application_index_scoped_metrics = list(_scoped_metrics)
_test_application_index_scoped_metrics.append(('Function/views:index', 1))

if DJANGO_VERSION >= (1, 5):
    _test_application_index_scoped_metrics.extend([
            ('Function/django.http.response:HttpResponse.close', 1)])


@validate_transaction_errors(errors=[])
@validate_transaction_metrics('views:index',
        scoped_metrics=_test_application_index_scoped_metrics)
@validate_code_level_metrics("views", "index")
def test_application_index(target_application):
    response = target_application.get('')
    response.mustcontain('INDEX RESPONSE')


_test_application_view_scoped_metrics = list(_scoped_metrics)
_test_application_view_scoped_metrics.append(('Function/urls:View.get', 1))

if DJANGO_VERSION >= (1, 5):
    _test_application_view_scoped_metrics.extend([
            ('Function/rest_framework.response:Response.close', 1)])


@validate_transaction_errors(errors=[])
@validate_transaction_metrics('urls:View.get',
    scoped_metrics=_test_application_view_scoped_metrics)
@validate_code_level_metrics("urls.View", "get")
def test_application_view(target_application):
    response = target_application.get('/view/')
    assert response.status_int == 200
    response.mustcontain('restframework view response')


_test_application_view_error_scoped_metrics = list(_scoped_metrics)
_test_application_view_error_scoped_metrics.append(
        ('Function/urls:ViewError.get', 1))


@validate_transaction_errors(errors=['urls:Error'])
@validate_transaction_metrics('urls:ViewError.get',
    scoped_metrics=_test_application_view_error_scoped_metrics)
@validate_code_level_metrics("urls.ViewError", "get")
def test_application_view_error(target_application):
    target_application.get('/view_error/', status=500)


_test_application_view_handle_error_scoped_metrics = list(_scoped_metrics)
_test_application_view_handle_error_scoped_metrics.append(
        ('Function/urls:ViewHandleError.get', 1))


@pytest.mark.parametrize('status,should_record', [(418, True), (200, False)])
@pytest.mark.parametrize('use_global_exc_handler', [True, False])
@validate_code_level_metrics("urls.ViewHandleError", "get")
def test_application_view_handle_error(status, should_record,
        use_global_exc_handler, target_application):
    errors = ['urls:Error'] if should_record else []

    @validate_transaction_errors(errors=errors)
    @validate_transaction_metrics('urls:ViewHandleError.get',
        scoped_metrics=_test_application_view_handle_error_scoped_metrics)
    def _test():
        response = target_application.get(
                '/view_handle_error/%s/%s/' % (status, use_global_exc_handler),
                status=status)
        if use_global_exc_handler:
            response.mustcontain('exception was handled global')
        else:
            response.mustcontain('exception was handled not global')

    _test()


_test_api_view_view_name_get = 'urls:wrapped_view.get'
_test_api_view_scoped_metrics_get = list(_scoped_metrics)
_test_api_view_scoped_metrics_get.append(
        ('Function/%s' % _test_api_view_view_name_get, 1))


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(_test_api_view_view_name_get,
    scoped_metrics=_test_api_view_scoped_metrics_get)
@validate_code_level_metrics("urls.WrappedAPIView" if six.PY3 else "urls", "wrapped_view")
def test_api_view_get(target_application):
    response = target_application.get('/api_view/')
    response.mustcontain('wrapped_view response')


_test_api_view_view_name_post = 'urls:wrapped_view.http_method_not_allowed'
_test_api_view_scoped_metrics_post = list(_scoped_metrics)
_test_api_view_scoped_metrics_post.append(
        ('Function/%s' % _test_api_view_view_name_post, 1))


@validate_transaction_errors(
        errors=['rest_framework.exceptions:MethodNotAllowed'])
@validate_transaction_metrics(_test_api_view_view_name_post,
    scoped_metrics=_test_api_view_scoped_metrics_post)
def test_api_view_method_not_allowed(target_application):
    target_application.post('/api_view/', status=405)


def test_application_view_agent_disabled(target_application):
    settings = global_settings()

    @override_generic_settings(settings, {'enabled': False})
    @function_not_called('newrelic.core.stats_engine',
            'StatsEngine.record_transaction')
    def _test():
        response = target_application.get('/view/')
        assert response.status_int == 200
        response.mustcontain('restframework view response')

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import random

# import re
import string

import pytest
from testing_support.fixtures import (
    make_cross_agent_headers,
    override_application_settings,
    validate_analytics_catmap_data,
)
from testing_support.validators.validate_transaction_event_attributes import (
    validate_transaction_event_attributes,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.application import application_instance
from newrelic.api.external_trace import ExternalTrace
from newrelic.api.transaction import Transaction
from newrelic.common.encoding_utils import deobfuscate

BASE_METRICS = [
    ("Function/_target_application:index", 1),
]
DT_METRICS = [
    ("Supportability/DistributedTrace/AcceptPayload/Success", None),
    ("Supportability/TraceContext/TraceParent/Accept/Success", 1),
]
BASE_ATTRS = ["response.status", "response.headers.contentType", "response.headers.contentLength"]


def raw_headers(response):
    try:
        # Manually encode into bytes
        return " ".join("%s: %s" % (k, v) for k, v in response.processed_headers).encode()
    except AttributeError:
        try:
            return response.get_headers()
        except AttributeError:
            return response.output()


@validate_transaction_metrics(
    "_target_application:index",
    scoped_metrics=BASE_METRICS,
    rollup_metrics=BASE_METRICS + DT_METRICS,
)
@override_application_settings(
    {
        "distributed_tracing.enabled": True,
    }
)
@validate_transaction_event_attributes(
    required_params={"agent": BASE_ATTRS, "user": [], "intrinsic": []},
)
def test_inbound_distributed_trace(app):
    transaction = Transaction(application_instance())
    dt_headers = ExternalTrace.generate_request_headers(transaction)

    response = app.fetch("get", "/", headers=dict(dt_headers))
    assert response.status == 200


ENCODING_KEY = "".join(random.choice(string.ascii_lowercase) for _ in range(40))  # nosec
_cat_response_header_urls_to_test = (
    ("/", "_target_application:index"),
    ("/streaming", "_target_application:streaming"),
    ("/error", "_target_application:error"),
)
_custom_settings = {
    "cross_process_id": "1#1",
    "encoding_key": ENCODING_KEY,
    "trusted_account_ids": [1],
    "cross_application_tracer.enabled": True,
    "distributed_tracing.enabled": False,
}


@pytest.mark.parametrize(
    "inbound_payload,expected_intrinsics,forgone_intrinsics,cat_id",
    [
        # Valid payload from trusted account
        (
            ["b854df4feb2b1f06", False, "7e249074f277923d", "5d2957be"],
            {
                "nr.referringTransactionGuid": "b854df4feb2b1f06",
                "nr.tripId": "7e249074f277923d",
                "nr.referringPathHash": "5d2957be",
            },
            [],
            "1#1",
        ),
        # Valid payload from an untrusted account
        (
            ["b854df4feb2b1f06", False, "7e249074f277923d", "5d2957be"],
            {},
            ["nr.referringTransactionGuid", "nr.tripId", "nr.referringPathHash"],
            "80#1",
        ),
    ],
)
@pytest.mark.parametrize("url,metric_name", _cat_response_header_urls_to_test)
def test_cat_response_headers(app, inbound_payload, expected_intrinsics, forgone_intrinsics, cat_id, url, metric_name):

    _base_metrics = [
        ("Function/%s" % metric_name, 1),
    ]

    @validate_transaction_metrics(
        metric_name,
        scoped_metrics=_base_metrics,
        rollup_metrics=_base_metrics,
    )
    @validate_analytics_catmap_data(
        "WebTransaction/Function/%s" % metric_name,
        expected_attributes=expected_intrinsics,
        non_expected_attributes=forgone_intrinsics,
    )
    @override_application_settings(_custom_settings)
    def _test():
        cat_headers = make_cross_agent_headers(inbound_payload, ENCODING_KEY, cat_id)
        response = app.fetch("get", url, headers=dict(cat_headers))
        if expected_intrinsics:
            # test valid CAT response header
            assert b"X-NewRelic-App-Data" in raw_headers(response)
            cat_response_header = response.headers.get("X-NewRelic-App-Data", None)

            app_data = json.loads(deobfuscate(cat_response_header, ENCODING_KEY))
            assert app_data[0] == cat_id
            assert app_data[1] == ("WebTransaction/Function/%s" % metric_name)
        else:
            assert b"X-NewRelic-App-Data" not in raw_headers(response)

    _test()


@override_application_settings(_custom_settings)
def test_cat_response_custom_header(app):
    inbound_payload = ["b854df4feb2b1f06", False, "7e249074f277923d", "5d2957be"]
    cat_id = "1#1"
    custom_header_value = b"my-custom-header-value"
    cat_headers = make_cross_agent_headers(inbound_payload, ENCODING_KEY, cat_id)

    response = app.fetch(
        "get", "/custom-header/%s/%s" % ("X-NewRelic-App-Data", custom_header_value), headers=dict(cat_headers)
    )
    assert custom_header_value in raw_headers(response), raw_headers(response)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

# Need to account for 4 possible variations of version declaration specified in (rejected) PEP 396
VERSION_ATTRS = ("__version__", "version", "__version_tuple__", "version_tuple")  # nosec
NULL_VERSIONS = frozenset((None, "", "0", "0.0", "0.0.0", "0.0.0.0", (0,), (0, 0), (0, 0, 0), (0, 0, 0, 0)))  # nosec


def get_package_version(name):
    """Gets the version string of the library.
    :param name: The name of library.
    :type name: str
    :return: The version of the library. Returns None if can't determine version.
    :type return: str or None

    Usage::
        >>> get_package_version("botocore")
                "1.1.0"
    """

    version = _get_package_version(name)

    # Coerce iterables into a string
    if isinstance(version, tuple):
        version = ".".join(str(v) for v in version)

    return version


def get_package_version_tuple(name):
    """Gets the version tuple of the library.
    :param name: The name of library.
    :type name: str
    :return: The version of the library. Returns None if can't determine version.
    :type return: tuple or None

    Usage::
        >>> get_package_version_tuple("botocore")
                (1, 1, 0)
    """

    def int_or_str(value):
        try:
            return int(value)
        except Exception:
            return str(value)

    version = _get_package_version(name)

    # Split "." separated strings and cast fields to ints
    if isinstance(version, str):
        version = tuple(int_or_str(v) for v in version.split("."))

    return version


def _get_package_version(name):
    module = sys.modules.get(name, None)
    version = None
    for attr in VERSION_ATTRS:
        try:
            version = getattr(module, attr, None)
            # Cast any version specified as a list into a tuple.
            version = tuple(version) if isinstance(version, list) else version
            if version not in NULL_VERSIONS:
                return version
        except Exception:
            pass

    # importlib was introduced into the standard library starting in Python3.8.
    if "importlib" in sys.modules and hasattr(sys.modules["importlib"], "metadata"):
        try:
            version = sys.modules["importlib"].metadata.version(name)  # pylint: disable=E1101
            if version not in NULL_VERSIONS:
                return version
        except Exception:
            pass

    if "pkg_resources" in sys.modules:
        try:
            version = sys.modules["pkg_resources"].get_distribution(name).version
            if version not in NULL_VERSIONS:
                return version
        except Exception:
            pass


# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

import pika
import pytest
from testing_support.db_settings import rabbitmq_settings
from testing_support.fixtures import (  # noqa: F401
    code_coverage_fixture,
    collector_agent_registration_fixture,
    collector_available_fixture,
)

QUEUE = "test_pika-%s" % uuid.uuid4()
QUEUE_2 = "test_pika-%s" % uuid.uuid4()

EXCHANGE = "exchange-%s" % uuid.uuid4()
EXCHANGE_2 = "exchange-%s" % uuid.uuid4()

CORRELATION_ID = "test-correlation-id"
REPLY_TO = "test-reply-to"
HEADERS = {"TestHeader": "my test header value"}
BODY = b"test_body"

DB_SETTINGS = rabbitmq_settings()[0]

_coverage_source = [
    "newrelic.hooks.messagebroker_pika",
]

code_coverage = code_coverage_fixture(source=_coverage_source)

_default_settings = {
    "transaction_tracer.explain_threshold": 0.0,
    "transaction_tracer.transaction_threshold": 0.0,
    "transaction_tracer.stack_trace_threshold": 0.0,
    "debug.log_data_collector_payloads": True,
    "debug.record_transaction_failure": True,
}

collector_agent_registration = collector_agent_registration_fixture(
    app_name="Python Agent Test (messagebroker_pika)",
    default_settings=_default_settings,
    linked_applications=["Python Agent Test (messagebroker_pika)"],
)


@pytest.fixture()
def producer():
    # put something into the queue so it can be consumed
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()

        channel.queue_declare(queue=QUEUE, durable=False)
        channel.exchange_declare(exchange=EXCHANGE, durable=False)
        channel.queue_bind(queue=QUEUE, exchange=EXCHANGE)

        channel.basic_publish(
            exchange=EXCHANGE,
            routing_key=QUEUE,
            body=BODY,
            properties=pika.spec.BasicProperties(correlation_id=CORRELATION_ID, reply_to=REPLY_TO, headers=HEADERS),
        )
        yield
        channel.queue_delete(queue=QUEUE)
        channel.exchange_delete(exchange=EXCHANGE)


@pytest.fixture()
def producer_2():
    # put something into the queue so it can be consumed
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()

        channel.queue_declare(queue=QUEUE_2, durable=False)
        channel.exchange_declare(exchange=EXCHANGE_2, durable=False)
        channel.queue_bind(queue=QUEUE_2, exchange=EXCHANGE_2)

        channel.basic_publish(
            exchange=EXCHANGE_2,
            routing_key=QUEUE_2,
            body=BODY,
            properties=pika.spec.BasicProperties(correlation_id=CORRELATION_ID, reply_to=REPLY_TO, headers=HEADERS),
        )
        yield
        channel.queue_delete(queue=QUEUE_2)
        channel.exchange_delete(exchange=EXCHANGE_2)


@pytest.fixture()
def produce_five():
    # put something into the queue so it can be consumed
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()

        channel.queue_declare(queue=QUEUE, durable=False)
        channel.exchange_declare(exchange=EXCHANGE, durable=False)
        channel.queue_bind(queue=QUEUE, exchange=EXCHANGE)

        for _ in range(5):
            channel.basic_publish(
                exchange=EXCHANGE,
                routing_key=QUEUE,
                body=BODY,
                properties=pika.spec.BasicProperties(correlation_id=CORRELATION_ID, reply_to=REPLY_TO, headers=HEADERS),
            )

        yield
        channel.queue_delete(queue=QUEUE)
        channel.exchange_delete(exchange=EXCHANGE)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import time

import webtest
from testing_support.fixtures import (
    cat_enabled,
    make_cross_agent_headers,
    make_synthetics_header,
    override_application_settings,
    reset_core_stats_engine,
    validate_error_event_sample_data,
    validate_transaction_error_event_count,
)
from testing_support.sample_applications import fully_featured_app
from testing_support.validators.validate_non_transaction_error_event import (
    validate_non_transaction_error_event,
)

from newrelic.api.application import application_instance as application
from newrelic.api.application import application_settings
from newrelic.api.time_trace import notice_error
from newrelic.common.object_names import callable_name

# Error in test app hard-coded as a ValueError
SYNTHETICS_RESOURCE_ID = "09845779-16ef-4fa7-b7f2-44da8e62931c"
SYNTHETICS_JOB_ID = "8c7dd3ba-4933-4cbb-b1ed-b62f511782f4"
SYNTHETICS_MONITOR_ID = "dc452ae9-1a93-4ab5-8a33-600521e9cd00"

ERR_MESSAGE = "Transaction had bad value"
ERROR = ValueError(ERR_MESSAGE)

fully_featured_application = webtest.TestApp(fully_featured_app)

_intrinsic_attributes = {
    "error.class": callable_name(ERROR),
    "error.message": ERR_MESSAGE,
    "error.expected": False,
    "transactionName": "WebTransaction/Uri/",
    "port": 80,  # SERVER_PORT default value in webtest WSGI environ
}


@validate_error_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs=False)
@validate_transaction_error_event_count(num_errors=1)
def test_transaction_error_event_no_extra_attributes():
    test_environ = {"err_message": ERR_MESSAGE, "record_attributes": "FALSE"}
    response = fully_featured_application.get("/", extra_environ=test_environ)


_intrinsic_attributes = {
    "error.class": callable_name(ERROR),
    "error.message": ERR_MESSAGE,
    "error.expected": False,
    "transactionName": "WebTransaction/Uri/",
    "databaseCallCount": 2,
    "externalCallCount": 2,
    "queueDuration": True,
    "port": 8888,
}


@validate_error_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs=True)
def test_transaction_error_event_lotsa_attributes():
    test_environ = {
        "err_message": ERR_MESSAGE,
        "external": "2",
        "db": "2",
        "mod_wsgi.queue_start": ("t=%r" % time.time()),
        "SERVER_PORT": "8888",
    }
    response = fully_featured_application.get("/", extra_environ=test_environ)


_intrinsic_attributes = {
    "error.class": callable_name(ERROR),
    "error.message": ERR_MESSAGE,
    "error.expected": False,
    "transactionName": "OtherTransaction/Uri/",
    "databaseCallCount": 2,
    "externalCallCount": 2,
    "queueDuration": False,
}


@validate_error_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs=True)
def test_transaction_error_background_task():
    test_environ = {"err_message": ERR_MESSAGE, "external": "2", "db": "2", "newrelic.set_background_task": True}
    response = fully_featured_application.get("/", extra_environ=test_environ)


_intrinsic_attributes = {
    "error.class": callable_name(ERROR),
    "error.message": ERR_MESSAGE,
    "error.expected": False,
    "transactionName": "WebTransaction/Uri/",
    "nr.referringTransactionGuid": 7,
}


@cat_enabled
@validate_error_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs=True)
def test_transaction_error_cross_agent():
    test_environ = {
        "err_message": ERR_MESSAGE,
    }
    settings = application_settings()
    transaction_data = [7, 1, 77, "/path-hash"]
    headers = make_cross_agent_headers(transaction_data, settings.encoding_key, settings.cross_process_id)
    response = fully_featured_application.get("/", headers=headers, extra_environ=test_environ)


_intrinsic_attributes = {
    "error.class": callable_name(ERROR),
    "error.message": ERR_MESSAGE,
    "error.expected": False,
    "transactionName": "WebTransaction/Uri/",
    "nr.syntheticsResourceId": SYNTHETICS_RESOURCE_ID,
    "nr.syntheticsJobId": SYNTHETICS_JOB_ID,
    "nr.syntheticsMonitorId": SYNTHETICS_MONITOR_ID,
}


@validate_error_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs=True)
def test_transaction_error_with_synthetics():
    test_environ = {
        "err_message": ERR_MESSAGE,
    }
    settings = application_settings()
    headers = make_synthetics_header(
        settings.trusted_account_ids[0],
        SYNTHETICS_RESOURCE_ID,
        SYNTHETICS_JOB_ID,
        SYNTHETICS_MONITOR_ID,
        settings.encoding_key,
    )
    response = fully_featured_application.get("/", headers=headers, extra_environ=test_environ)


_intrinsic_attributes = {
    "error.class": callable_name(ERROR),
    "error.message": ERR_MESSAGE,
    "error.expected": False,
    "transactionName": "WebTransaction/Uri/",
}


@validate_error_event_sample_data(required_attrs=_intrinsic_attributes, required_user_attrs=True, num_errors=2)
@validate_transaction_error_event_count(num_errors=2)
def test_multiple_errors_in_transaction():
    test_environ = {
        "err_message": ERR_MESSAGE,
        "n_errors": "2",
    }
    response = fully_featured_application.get("/", extra_environ=test_environ)


@override_application_settings({"error_collector.enabled": False})
@validate_error_event_sample_data(num_errors=0)
@validate_transaction_error_event_count(num_errors=0)
def test_error_collector_disabled():
    """If error_collector is disabled, don't event collect error info. There
    should be an empty result from transaction_node.error_event"""
    test_environ = {
        "err_message": ERR_MESSAGE,
    }
    response = fully_featured_application.get("/", extra_environ=test_environ)


@override_application_settings({"collect_error_events": False})
@validate_transaction_error_event_count(num_errors=0)
def test_collect_error_events_false():
    """Don't save error events to stats engine. Error info can be collected
    for error traces, so we only validate that event is not on stats engine,
    not that it can't be generated from the transaction node."""
    test_environ = {
        "err_message": ERR_MESSAGE,
    }
    response = fully_featured_application.get("/", extra_environ=test_environ)


@override_application_settings({"error_collector.capture_events": False})
@validate_transaction_error_event_count(num_errors=0)
def test_collect_error_capture_events_disabled():
    """Don't save error events to stats engine. Error info can be collected
    for error traces, so we only validate that event is not on stats engine,
    not that it can't be generated from the transaction node."""
    test_environ = {
        "err_message": ERR_MESSAGE,
    }
    response = fully_featured_application.get("/", extra_environ=test_environ)


# -------------- Test Error Events outside of transaction ----------------


class ErrorEventOutsideTransactionError(Exception):
    pass


outside_error = ErrorEventOutsideTransactionError(ERR_MESSAGE)

_intrinsic_attributes = {
    "error.class": callable_name(outside_error),
    "error.message": ERR_MESSAGE,
    "error.expected": False,
}


@reset_core_stats_engine()
@validate_non_transaction_error_event(_intrinsic_attributes)
def test_error_event_outside_transaction():
    try:
        raise outside_error
    except ErrorEventOutsideTransactionError:
        app = application()
        notice_error(sys.exc_info(), application=app)


_err_params = {"key": "value"}


@reset_core_stats_engine()
@validate_non_transaction_error_event(_intrinsic_attributes, required_user=_err_params)
def test_error_event_with_params_outside_transaction():
    try:
        raise outside_error
    except ErrorEventOutsideTransactionError:
        app = application()
        notice_error(sys.exc_info(), attributes=_err_params, application=app)


@reset_core_stats_engine()
@validate_non_transaction_error_event(_intrinsic_attributes, num_errors=2)
def test_multiple_error_events_outside_transaction():
    app = application()
    for i in range(2):
        try:
            raise ErrorEventOutsideTransactionError(ERR_MESSAGE + str(i))
        except ErrorEventOutsideTransactionError:
            notice_error(sys.exc_info(), application=app)


@reset_core_stats_engine()
@override_application_settings({"error_collector.enabled": False})
@validate_non_transaction_error_event(num_errors=0)
def test_error_event_outside_transaction_error_collector_disabled():
    try:
        raise outside_error
    except ErrorEventOutsideTransactionError:
        app = application()
        notice_error(sys.exc_info(), application=app)


@reset_core_stats_engine()
@override_application_settings({"error_collector.capture_events": False})
@validate_non_transaction_error_event(num_errors=0)
def test_error_event_outside_transaction_capture_events_disabled():
    try:
        raise outside_error
    except ErrorEventOutsideTransactionError:
        app = application()
        notice_error(sys.exc_info(), application=app)


@reset_core_stats_engine()
@override_application_settings({"collect_error_events": False})
@validate_non_transaction_error_event(num_errors=0)
def test_error_event_outside_transaction_collect_error_events_false():
    try:
        raise outside_error
    except ErrorEventOutsideTransactionError:
        app = application()
        notice_error(sys.exc_info(), application=app)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import six
from testing_support.fixtures import (
    code_coverage_fixture,
    collector_agent_registration_fixture,
    collector_available_fixture,
)

_coverage_source = [
    "newrelic.hooks.framework_graphql",
]

code_coverage = code_coverage_fixture(source=_coverage_source)

_default_settings = {
    "transaction_tracer.explain_threshold": 0.0,
    "transaction_tracer.transaction_threshold": 0.0,
    "transaction_tracer.stack_trace_threshold": 0.0,
    "debug.log_data_collector_payloads": True,
    "debug.record_transaction_failure": True,
}

collector_agent_registration = collector_agent_registration_fixture(
    app_name="Python Agent Test (framework_graphql)",
    default_settings=_default_settings,
)


@pytest.fixture(scope="session")
def app():
    from _target_application import _target_application

    return _target_application


if six.PY2:
    collect_ignore = ["test_application_async.py"]



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import tempfile
import time

import pytest
from testing_support.fixtures import (
    failing_endpoint,
    function_not_called,
    override_generic_settings,
)

from newrelic.common.agent_http import DeveloperModeClient
from newrelic.common.object_wrapper import function_wrapper, transient_function_wrapper
from newrelic.core.application import Application
from newrelic.core.config import finalize_application_settings, global_settings
from newrelic.core.custom_event import create_custom_event
from newrelic.core.error_node import ErrorNode
from newrelic.core.function_node import FunctionNode
from newrelic.core.log_event_node import LogEventNode
from newrelic.core.root_node import RootNode
from newrelic.core.stats_engine import CustomMetrics, SampledDataSet
from newrelic.core.transaction_node import TransactionNode
from newrelic.network.exceptions import RetryDataForRequest

settings = global_settings()


@pytest.fixture(scope="module")
def transaction_node(request):
    default_capacity = SampledDataSet().capacity
    num_events = default_capacity + 1

    custom_events = SampledDataSet(capacity=num_events)
    for _ in range(num_events):
        event = create_custom_event("Custom", {})
        custom_events.add(event)

    log_events = SampledDataSet(capacity=num_events)
    for _ in range(num_events):
        event = LogEventNode(1653609717, "WARNING", "A", {})
        log_events.add(event)

    error = ErrorNode(
        timestamp=0,
        type="foo:bar",
        message="oh no! your foo had a bar",
        expected=False,
        span_id=None,
        stack_trace="",
        custom_params={},
        file_name=None,
        line_number=None,
        source=None,
    )

    errors = tuple(error for _ in range(num_events))

    function = FunctionNode(
        group="Function",
        name="foo",
        children=(),
        start_time=0,
        end_time=1,
        duration=1,
        exclusive=1,
        label=None,
        params=None,
        rollup=None,
        guid="GUID",
        agent_attributes={},
        user_attributes={},
    )

    children = tuple(function for _ in range(num_events))

    root = RootNode(
        name="Function/main",
        children=children,
        start_time=1524764430.0,
        end_time=1524764430.1,
        duration=0.1,
        exclusive=0.1,
        guid=None,
        agent_attributes={},
        user_attributes={},
        path="OtherTransaction/Function/main",
        trusted_parent_span=None,
        tracing_vendors=None,
    )

    node = TransactionNode(
        settings=finalize_application_settings({"agent_run_id": "1234567"}),
        path="OtherTransaction/Function/main",
        type="OtherTransaction",
        group="Function",
        base_name="main",
        name_for_metric="Function/main",
        port=None,
        request_uri=None,
        queue_start=0.0,
        start_time=1524764430.0,
        end_time=1524764430.1,
        last_byte_time=0.0,
        total_time=0.1,
        response_time=0.1,
        duration=0.1,
        exclusive=0.1,
        root=root,
        errors=errors,
        slow_sql=(),
        custom_events=custom_events,
        log_events=log_events,
        apdex_t=0.5,
        suppress_apdex=False,
        custom_metrics=CustomMetrics(),
        guid="4485b89db608aece",
        cpu_time=0.0,
        suppress_transaction_trace=False,
        client_cross_process_id=None,
        referring_transaction_guid=None,
        record_tt=False,
        synthetics_resource_id=None,
        synthetics_job_id=None,
        synthetics_monitor_id=None,
        synthetics_header=None,
        is_part_of_cat=False,
        trip_id="4485b89db608aece",
        path_hash=None,
        referring_path_hash=None,
        alternate_path_hashes=[],
        trace_intrinsics={},
        distributed_trace_intrinsics={},
        agent_attributes=[],
        user_attributes=[],
        priority=1.0,
        parent_transport_duration=None,
        parent_span=None,
        parent_type=None,
        parent_account=None,
        parent_app=None,
        parent_tx=None,
        parent_transport_type=None,
        sampled=True,
        root_span_guid=None,
        trace_id="4485b89db608aece",
        loop_time=0.0,
    )
    return node


def validate_metric_payload(metrics=[], endpoints_called=[]):
    sent_metrics = {}

    @transient_function_wrapper("newrelic.core.agent_protocol", "AgentProtocol.send")
    def send_request_wrapper(wrapped, instance, args, kwargs):
        def _bind_params(method, payload=(), *args, **kwargs):
            return method, payload

        method, payload = _bind_params(*args, **kwargs)
        endpoints_called.append(method)

        if method == "metric_data" and payload:
            for metric_info, metric_values in payload[3]:
                metric_key = (metric_info["name"], metric_info["scope"])
                sent_metrics[metric_key] = metric_values

        return wrapped(*args, **kwargs)

    @function_wrapper
    def _wrapper(wrapped, instance, args, kwargs):
        def validate():
            for metric_name, count in metrics:
                metric_key = (metric_name, "")  # only search unscoped

                if count is not None:
                    assert metric_key in sent_metrics, metric_key
                    assert sent_metrics[metric_key][0] == count, metric_key
                else:
                    assert metric_key not in sent_metrics, metric_key

        _new_wrapper = send_request_wrapper(wrapped)
        val = _new_wrapper(*args, **kwargs)
        validate()
        return val

    return _wrapper


def validate_transaction_event_payloads(payload_validators):
    @function_wrapper
    def _wrapper(wrapped, instance, args, kwargs):
        payloads = []

        @transient_function_wrapper("newrelic.core.agent_protocol", "AgentProtocol.send")
        def send_request_wrapper(wrapped, instance, args, kwargs):
            def _bind_params(method, payload=(), *args, **kwargs):
                return method, payload

            method, payload = _bind_params(*args, **kwargs)

            if method == "analytic_event_data":
                payloads.append(payload)

            return wrapped(*args, **kwargs)

        _new_wrapper = send_request_wrapper(wrapped)
        val = _new_wrapper(*args, **kwargs)

        assert len(payloads) == len(payload_validators)

        for payload, validator in zip(payloads, payload_validators):
            validator(payload)

        return val

    return _wrapper


def validate_error_event_sampling(events_seen, reservoir_size, endpoints_called=[]):
    @transient_function_wrapper("newrelic.core.data_collector", "AgentProtocol.send")
    def send_request_wrapper(wrapped, instance, args, kwargs):
        def _bind_params(method, payload=(), *args, **kwargs):
            return method, payload

        method, payload = _bind_params(*args, **kwargs)
        endpoints_called.append(method)

        if method == "error_event_data":
            sampling_info = payload[1]
            assert sampling_info["events_seen"] == events_seen
            assert sampling_info["reservoir_size"] == reservoir_size

        return wrapped(*args, **kwargs)

    return send_request_wrapper


required_metrics = [
    ("Supportability/Events/TransactionError/Seen", 0),
    ("Supportability/Events/TransactionError/Sent", 0),
    ("Supportability/Events/Customer/Seen", 0),
    ("Supportability/Events/Customer/Sent", 0),
    ("Supportability/Python/RequestSampler/requests", 1),
    ("Supportability/Python/RequestSampler/samples", 1),
    ("Instance/Reporting", 1),
]

endpoints_called = []


@validate_metric_payload(metrics=required_metrics, endpoints_called=endpoints_called)
@override_generic_settings(
    settings,
    {
        "developer_mode": True,
        "license_key": "**NOT A LICENSE KEY**",
        "feature_flag": set(),
        "audit_log_file": tempfile.NamedTemporaryFile().name,
    },
)
def test_application_harvest():
    app = Application("Python Agent Test (Harvest Loop)")
    app.connect_to_data_collector(None)
    app.harvest()

    # Verify that the metric_data endpoint is the 2nd to last endpoint called
    # Last endpoint called is get_agent_commands
    assert endpoints_called[-2] == "metric_data"

    # verify audit log is not empty
    with open(settings.audit_log_file, "rb") as f:
        assert f.read()


@override_generic_settings(
    settings,
    {
        "serverless_mode.enabled": True,
        "license_key": "**NOT A LICENSE KEY**",
        "feature_flag": set(),
        "audit_log_file": tempfile.NamedTemporaryFile().name,
    },
)
def test_serverless_application_harvest():
    app = Application("Python Agent Test (Serverless Harvest Loop)")
    app.connect_to_data_collector(None)
    app.harvest()

    # verify audit log is not empty
    with open(settings.audit_log_file, "rb") as f:
        assert f.read()


@pytest.mark.parametrize(
    "distributed_tracing_enabled,span_events_enabled,spans_created",
    [
        (True, True, 1),
        (True, True, 15),
        (True, False, 1),
        (True, True, 0),
        (True, False, 0),
        (False, True, 0),
    ],
)
def test_application_harvest_with_spans(distributed_tracing_enabled, span_events_enabled, spans_created):
    span_endpoints_called = []
    max_samples_stored = 10

    if distributed_tracing_enabled and span_events_enabled:
        seen = spans_created
        sent = min(spans_created, max_samples_stored)
    else:
        seen = None
        sent = None

    spans_required_metrics = list(required_metrics)

    spans_required_metrics.extend(
        [
            ("Supportability/SpanEvent/TotalEventsSeen", seen),
            ("Supportability/SpanEvent/TotalEventsSent", sent),
        ]
    )

    @validate_metric_payload(metrics=spans_required_metrics, endpoints_called=span_endpoints_called)
    @override_generic_settings(
        settings,
        {
            "developer_mode": True,
            "license_key": "**NOT A LICENSE KEY**",
            "distributed_tracing.enabled": distributed_tracing_enabled,
            "span_events.enabled": span_events_enabled,
            "event_harvest_config.harvest_limits.span_event_data": max_samples_stored,
        },
    )
    def _test():
        app = Application("Python Agent Test (Harvest Loop)")
        app.connect_to_data_collector(None)

        for _ in range(spans_created):
            app._stats_engine.span_events.add("event")

        assert app._stats_engine.span_events.num_samples == (min(spans_created, max_samples_stored))
        app.harvest()
        assert app._stats_engine.span_events.num_samples == 0

        # Verify that the metric_data endpoint is the 2nd to last and
        # span_event_data is the 3rd to last endpoint called
        assert span_endpoints_called[-2] == "metric_data"

        if span_events_enabled and spans_created > 0:
            assert span_endpoints_called[-3] == "span_event_data"
        else:
            assert span_endpoints_called[-3] != "span_event_data"

    _test()


@pytest.mark.parametrize(
    "span_queue_size, spans_to_send, expected_seen, expected_sent",
    (
        (0, 1, 1, 0),
        (1, 1, 1, 1),
        (7, 10, 10, 7),
    ),
)
@pytest.mark.parametrize("span_batching", (True, False))
def test_application_harvest_with_span_streaming(
    span_batching, span_queue_size, spans_to_send, expected_sent, expected_seen
):
    @override_generic_settings(
        settings,
        {
            "developer_mode": True,
            "distributed_tracing.enabled": True,
            "span_events.enabled": True,
            "infinite_tracing._trace_observer_host": "x",
            "infinite_tracing.span_queue_size": span_queue_size,
            "infinite_tracing.batching": span_batching,
        },
    )
    @validate_metric_payload(
        metrics=[
            ("Supportability/InfiniteTracing/Span/Seen", expected_seen),
            ("Supportability/InfiniteTracing/Span/Sent", expected_sent),
        ],
        endpoints_called=[],
    )
    def _test():
        app = Application("Python Agent Test (Harvest Loop)")
        app.connect_to_data_collector(None)

        for _ in range(spans_to_send):
            app._stats_engine.span_stream.put(None)
        app.harvest()

    _test()


@failing_endpoint("metric_data")
@pytest.mark.parametrize("span_events_enabled", (True, False))
def test_failed_spans_harvest(span_events_enabled):
    @override_generic_settings(
        settings,
        {
            "developer_mode": True,
            "license_key": "**NOT A LICENSE KEY**",
            "distributed_tracing.enabled": True,
            "span_events.enabled": span_events_enabled,
        },
    )
    def _test():
        # Test that if an endpoint call that occurs after we successfully send
        # span data fails, we do not try to send span data again with the next
        # harvest.

        app = Application("Python Agent Test (Harvest Loop)")
        app.connect_to_data_collector(None)

        app._stats_engine.span_events.add("event")
        assert app._stats_engine.span_events.num_samples == 1
        app.harvest()
        assert app._stats_engine.span_events.num_samples == 0

    _test()


@override_generic_settings(
    settings,
    {
        "developer_mode": True,
        "license_key": "**NOT A LICENSE KEY**",
        "feature_flag": set(),
        "collect_custom_events": False,
        "application_logging.forwarding.enabled": False,
    },
)
def test_transaction_count(transaction_node):
    app = Application("Python Agent Test (Harvest Loop)")
    app.connect_to_data_collector(None)

    app.record_transaction(transaction_node)

    # Harvest has not run yet
    assert app._transaction_count == 1

    app.harvest()

    # Harvest resets the transaction count
    assert app._transaction_count == 0

    # Record a transaction
    app.record_transaction(transaction_node)
    assert app._transaction_count == 1

    app.harvest()

    # Harvest resets the transaction count
    assert app._transaction_count == 0


@override_generic_settings(
    settings,
    {
        "developer_mode": True,
        "license_key": "**NOT A LICENSE KEY**",
        "feature_flag": set(),
    },
)
def test_adaptive_sampling(transaction_node, monkeypatch):
    app = Application("Python Agent Test (Harvest Loop)")

    # Should always return false for sampling prior to connect
    assert app.compute_sampled() is False

    app.connect_to_data_collector(None)

    # First harvest, first N should be sampled
    for _ in range(settings.sampling_target):
        assert app.compute_sampled() is True

    assert app.compute_sampled() is False

    # fix random.randrange to return 0
    monkeypatch.setattr(random, "randrange", lambda *args, **kwargs: 0)

    # Multiple resets should behave the same
    for _ in range(2):
        # Set the last_reset to longer than the period so a reset will occur.
        app.adaptive_sampler.last_reset = time.time() - app.adaptive_sampler.period

        # Subsequent harvests should allow sampling of 2X the target
        for _ in range(2 * settings.sampling_target):
            assert app.compute_sampled() is True

        # No further samples should be saved
        assert app.compute_sampled() is False


@override_generic_settings(
    settings,
    {
        "developer_mode": True,
        "license_key": "**NOT A LICENSE KEY**",
        "feature_flag": set(),
        "distributed_tracing.enabled": True,
        "application_logging.forwarding.enabled": True,
        "event_harvest_config.harvest_limits.error_event_data": 1000,
        "event_harvest_config.harvest_limits.span_event_data": 1000,
        "event_harvest_config.harvest_limits.custom_event_data": 1000,
        "event_harvest_config.harvest_limits.log_event_data": 1000,
    },
)
def test_reservoir_sizes(transaction_node):
    app = Application("Python Agent Test (Harvest Loop)")
    app.connect_to_data_collector(None)

    # Record a transaction with events
    app.record_transaction(transaction_node)

    # Test that the samples have been recorded
    assert app._stats_engine.custom_events.num_samples == 101
    assert app._stats_engine.error_events.num_samples == 101
    assert app._stats_engine.log_events.num_samples == 101

    # Add 1 for the root span
    assert app._stats_engine.span_events.num_samples == 102


@pytest.mark.parametrize(
    "harvest_name, event_name",
    [
        ("analytic_event_data", "transaction_events"),
        ("error_event_data", "error_events"),
        ("custom_event_data", "custom_events"),
        ("log_event_data", "log_events"),
        ("span_event_data", "span_events"),
    ],
)
@override_generic_settings(
    settings,
    {
        "developer_mode": True,
        "license_key": "**NOT A LICENSE KEY**",
        "feature_flag": set(),
        "distributed_tracing.enabled": True,
    },
)
def test_reservoir_size_zeros(harvest_name, event_name):
    app = Application("Python Agent Test (Harvest Loop)")
    app.connect_to_data_collector(None)

    setattr(settings.event_harvest_config.harvest_limits, harvest_name, 0)
    settings.event_harvest_config.allowlist = frozenset(())
    app._stats_engine.reset_stats(settings)

    app._stats_engine.transaction_events.add("transaction event")
    app._stats_engine.error_events.add("error event")
    app._stats_engine.custom_events.add("custom event")
    app._stats_engine.log_events.add(LogEventNode(1653609717, "WARNING", "A", {}))
    app._stats_engine.span_events.add("span event")

    assert app._stats_engine.transaction_events.num_seen == 1
    assert app._stats_engine.error_events.num_seen == 1
    assert app._stats_engine.custom_events.num_seen == 1
    assert app._stats_engine.log_events.num_seen == 1
    assert app._stats_engine.span_events.num_seen == 1

    stat_events = set(("transaction_events", "error_events", "custom_events", "log_events", "span_events"))

    for stat_event in stat_events:
        event = getattr(app._stats_engine, stat_event)

        if stat_event == event_name:
            assert event.num_samples == 0
        else:
            assert event.num_samples == 1

    app.harvest()

    assert app._stats_engine.transaction_events.num_seen == 0
    assert app._stats_engine.error_events.num_seen == 0
    assert app._stats_engine.custom_events.num_seen == 0
    assert app._stats_engine.log_events.num_seen == 0
    assert app._stats_engine.span_events.num_seen == 0


@pytest.mark.parametrize("events_seen", (1, 5, 10))
def test_error_event_sampling_info(events_seen):
    reservoir_size = 5
    endpoints_called = []

    @validate_error_event_sampling(
        events_seen=events_seen, reservoir_size=reservoir_size, endpoints_called=endpoints_called
    )
    @override_generic_settings(
        settings,
        {
            "developer_mode": True,
            "license_key": "**NOT A LICENSE KEY**",
            "event_harvest_config.harvest_limits.error_event_data": reservoir_size,
        },
    )
    def _test():
        app = Application("Python Agent Test (Harvest Loop)")
        app.connect_to_data_collector(None)

        for _ in range(events_seen):
            app._stats_engine.error_events.add("error")

        app.harvest()

    _test()
    assert "error_event_data" in endpoints_called


@pytest.mark.parametrize(
    "time_to_next_reset,computed_count,computed_count_last",
    (
        (20, 124, 10),  # no reset
        (-8, 1, 123),  # reset to last transaction count
        (-68, 1, 10),  # more than 1 minute passed, reset fully
    ),
)
@override_generic_settings(
    settings,
    {
        "serverless_mode.enabled": True,
    },
)
def test_serverless_mode_adaptive_sampling(time_to_next_reset, computed_count, computed_count_last, monkeypatch):
    # fix random.randrange to return 0
    monkeypatch.setattr(random, "randrange", lambda *args, **kwargs: 0)

    app = Application("Python Agent Test (Harvest Loop)")

    app.connect_to_data_collector(None)
    app.adaptive_sampler.computed_count = 123
    app.adaptive_sampler.last_reset = time.time() - 60 + time_to_next_reset

    assert app.compute_sampled() is True
    assert app.adaptive_sampler.computed_count == computed_count
    assert app.adaptive_sampler.computed_count_last == computed_count_last


@function_not_called("newrelic.core.adaptive_sampler", "AdaptiveSampler._reset")
@override_generic_settings(
    settings,
    {
        "developer_mode": True,
    },
)
def test_compute_sampled_no_reset():
    app = Application("Python Agent Test (Harvest Loop)")
    app.connect_to_data_collector(None)
    assert app.compute_sampled() is True


def test_analytic_event_sampling_info():
    synthetics_limit = 10
    transactions_limit = 20

    def synthetics_validator(payload):
        _, sampling_info, _ = payload
        assert sampling_info["reservoir_size"] == synthetics_limit
        assert sampling_info["events_seen"] == 1

    def transactions_validator(payload):
        _, sampling_info, _ = payload
        assert sampling_info["reservoir_size"] == transactions_limit
        assert sampling_info["events_seen"] == 1

    validators = [synthetics_validator, transactions_validator]

    @validate_transaction_event_payloads(validators)
    @override_generic_settings(
        settings,
        {
            "developer_mode": True,
            "event_harvest_config.harvest_limits.analytic_event_data": transactions_limit,
            "agent_limits.synthetics_events": synthetics_limit,
        },
    )
    def _test():
        app = Application("Python Agent Test (Harvest Loop)")
        app.connect_to_data_collector(None)

        app._stats_engine.transaction_events.add("transaction event")
        app._stats_engine.synthetics_events.add("synthetic event")

        app.harvest()

    _test()


@pytest.mark.parametrize("has_synthetic_events", (True, False))
@pytest.mark.parametrize("has_transaction_events", (True, False))
@override_generic_settings(
    settings,
    {
        "developer_mode": True,
    },
)
def test_analytic_event_payloads(has_synthetic_events, has_transaction_events):
    def synthetics_validator(payload):
        events = payload[-1]
        assert list(events) == ["synthetic event"]

    def transactions_validator(payload):
        events = payload[-1]
        assert list(events) == ["transaction event"]

    validators = []
    if has_synthetic_events:
        validators.append(synthetics_validator)
    if has_transaction_events:
        validators.append(transactions_validator)

    @validate_transaction_event_payloads(validators)
    def _test():
        app = Application("Python Agent Test (Harvest Loop)")
        app.connect_to_data_collector(None)

        if has_transaction_events:
            app._stats_engine.transaction_events.add("transaction event")

        if has_synthetic_events:
            app._stats_engine.synthetics_events.add("synthetic event")

        app.harvest()

    _test()


@override_generic_settings(
    settings,
    {
        "developer_mode": True,
        "collect_analytics_events": False,
        "transaction_events.enabled": False,
    },
)
def test_transaction_events_disabled():
    endpoints_called = []
    expected_metrics = (
        ("Supportability/Python/RequestSampler/requests", None),
        ("Supportability/Python/RequestSampler/samples", None),
    )

    @validate_metric_payload(expected_metrics, endpoints_called)
    def _test():
        app = Application("Python Agent Test (Harvest Loop)")
        app.connect_to_data_collector(None)
        app.harvest()

    _test()
    assert "metric_data" in endpoints_called


@failing_endpoint("analytic_event_data", call_number=2)
@override_generic_settings(
    settings,
    {
        "developer_mode": True,
        "license_key": "**NOT A LICENSE KEY**",
    },
)
def test_reset_synthetics_events():
    app = Application("Python Agent Test (Harvest Loop)")
    app.connect_to_data_collector(None)

    app._stats_engine.synthetics_events.add("synthetics event")
    app._stats_engine.transaction_events.add("transaction event")

    assert app._stats_engine.synthetics_events.num_seen == 1
    assert app._stats_engine.transaction_events.num_seen == 1

    app.harvest()

    assert app._stats_engine.synthetics_events.num_seen == 0
    assert app._stats_engine.transaction_events.num_seen == 1


@pytest.mark.parametrize(
    "allowlist_event",
    ("analytic_event_data", "custom_event_data", "log_event_data", "error_event_data", "span_event_data"),
)
@override_generic_settings(
    settings,
    {
        "developer_mode": True,
        "license_key": "**NOT A LICENSE KEY**",
    },
)
def test_flexible_events_harvested(allowlist_event):
    app = Application("Python Agent Test (Harvest Loop)")
    app.connect_to_data_collector(None)

    settings.event_harvest_config.allowlist = frozenset((allowlist_event,))
    app._stats_engine.reset_stats(settings)

    app._stats_engine.transaction_events.add("transaction event")
    app._stats_engine.error_events.add("error event")
    app._stats_engine.custom_events.add("custom event")
    app._stats_engine.log_events.add(LogEventNode(1653609717, "WARNING", "A", {}))
    app._stats_engine.span_events.add("span event")
    app._stats_engine.record_custom_metric("CustomMetric/Int", 1)

    assert app._stats_engine.transaction_events.num_seen == 1
    assert app._stats_engine.error_events.num_seen == 1
    assert app._stats_engine.custom_events.num_seen == 1
    assert app._stats_engine.log_events.num_seen == 1
    assert app._stats_engine.span_events.num_seen == 1
    assert app._stats_engine.record_custom_metric("CustomMetric/Int", 1)

    app.harvest(flexible=True)

    num_seen = 0 if (allowlist_event == "analytic_event_data") else 1
    assert app._stats_engine.transaction_events.num_seen == num_seen

    num_seen = 0 if (allowlist_event == "error_event_data") else 1
    assert app._stats_engine.error_events.num_seen == num_seen

    num_seen = 0 if (allowlist_event == "custom_event_data") else 1
    assert app._stats_engine.custom_events.num_seen == num_seen

    num_seen = 0 if (allowlist_event == "log_event_data") else 1
    assert app._stats_engine.log_events.num_seen == num_seen

    num_seen = 0 if (allowlist_event == "span_event_data") else 1
    assert app._stats_engine.span_events.num_seen == num_seen

    assert ("CustomMetric/Int", "") in app._stats_engine.stats_table
    assert app._stats_engine.metrics_count() > 1


@pytest.mark.parametrize(
    "allowlist_event",
    ("analytic_event_data", "custom_event_data", "log_event_data", "error_event_data", "span_event_data"),
)
@override_generic_settings(
    settings,
    {
        "developer_mode": True,
        "license_key": "**NOT A LICENSE KEY**",
    },
)
def test_default_events_harvested(allowlist_event):
    app = Application("Python Agent Test (Harvest Loop)")
    app.connect_to_data_collector(None)

    settings.event_harvest_config.allowlist = frozenset((allowlist_event,))
    app._stats_engine.reset_stats(settings)

    app._stats_engine.transaction_events.add("transaction event")
    app._stats_engine.error_events.add("error event")
    app._stats_engine.custom_events.add("custom event")
    app._stats_engine.log_events.add(LogEventNode(1653609717, "WARNING", "A", {}))
    app._stats_engine.span_events.add("span event")

    assert app._stats_engine.transaction_events.num_seen == 1
    assert app._stats_engine.error_events.num_seen == 1
    assert app._stats_engine.custom_events.num_seen == 1
    assert app._stats_engine.log_events.num_seen == 1
    assert app._stats_engine.span_events.num_seen == 1
    assert app._stats_engine.metrics_count() == 0

    app.harvest(flexible=False)

    num_seen = 0 if (allowlist_event != "analytic_event_data") else 1
    assert app._stats_engine.transaction_events.num_seen == num_seen

    num_seen = 0 if (allowlist_event != "error_event_data") else 1
    assert app._stats_engine.error_events.num_seen == num_seen

    num_seen = 0 if (allowlist_event != "custom_event_data") else 1
    assert app._stats_engine.custom_events.num_seen == num_seen

    num_seen = 0 if (allowlist_event != "log_event_data") else 1
    assert app._stats_engine.log_events.num_seen == num_seen

    num_seen = 0 if (allowlist_event != "span_event_data") else 1
    assert app._stats_engine.span_events.num_seen == num_seen

    assert app._stats_engine.metrics_count() == 4


@failing_endpoint("analytic_event_data")
@override_generic_settings(
    settings,
    {
        "developer_mode": True,
        "agent_limits.merge_stats_maximum": 0,
    },
)
def test_infinite_merges():
    app = Application("Python Agent Test (Harvest Loop)")
    app.connect_to_data_collector(None)

    app._stats_engine.transaction_events.add("transaction event")

    assert app._stats_engine.transaction_events.num_seen == 1

    app.harvest()

    # the agent_limits.merge_stats_maximum is not respected
    assert app._stats_engine.transaction_events.num_seen == 1


@failing_endpoint("analytic_event_data")
@override_generic_settings(
    settings,
    {
        "developer_mode": True,
    },
)
def test_flexible_harvest_rollback():
    app = Application("Python Agent Test (Harvest Loop)")
    app.connect_to_data_collector(None)

    settings.event_harvest_config.allowlist = frozenset(("analytic_event_data",))
    app._stats_engine.reset_stats(settings)

    # Cause a transaction event to attempt sending
    app._stats_engine.transaction_events.add("Custom/test_flexible_harvest_rollback")

    # Add a metric to the stats engine
    app._stats_engine.record_custom_metric("Custom/test_flexible_harvest_rollback", 1)
    stats_key = ("Custom/test_flexible_harvest_rollback", "")
    assert stats_key in app._stats_engine.stats_table

    app.harvest(flexible=True)

    # The metric should not be changed after a rollback
    assert app._stats_engine.stats_table[stats_key].call_count == 1


@override_generic_settings(
    settings,
    {
        "developer_mode": True,
    },
)
def test_get_agent_commands_returns_none():
    MISSING = object()
    original_return_value = DeveloperModeClient.RESPONSES.get("get_agent_commands", MISSING)
    DeveloperModeClient.RESPONSES["get_agent_commands"] = None

    try:
        app = Application("Python Agent Test (Harvest Loop)")
        app.connect_to_data_collector(None)
        app.process_agent_commands()
    finally:
        if original_return_value is MISSING:
            DeveloperModeClient.RESPONSES.pop("get_agent_commands")
        else:
            DeveloperModeClient.RESPONSES["get_agent_commands"] = original_return_value


@failing_endpoint("get_agent_commands")
@override_generic_settings(
    settings,
    {
        "developer_mode": True,
    },
)
def test_get_agent_commands_raises():
    app = Application("Python Agent Test (Harvest Loop)")
    app.connect_to_data_collector(None)
    with pytest.raises(RetryDataForRequest):
        app.process_agent_commands()
